from typing import *

import attr

from dlms_cosem.cosem import selective_access

from .base import CosemAttribute


@attr.s(auto_attribs=True)
class CosemAttributeWithSelection:
    attribute: CosemAttribute
    access_selection: Optional[
        Union[selective_access.RangeDescriptor, selective_access.EntryDescriptor]
    ]

    @classmethod
    def from_bytes(cls, source_bytes: bytes) -> "CosemAttributeWithSelection":
        cosem_attribute_data = source_bytes[:9]
        cosem_attribute = CosemAttribute.from_bytes(cosem_attribute_data)
        data = bytearray(source_bytes[9:])
        has_access_selection = bool(data.pop(0))
        if has_access_selection:
            access_selection = selective_access.AccessDescriptorFactory.from_bytes(data)
        else:
            access_selection = None

        return cls(cosem_attribute, access_selection)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.extend(self.attribute.to_bytes())
        if self.access_selection:
            out.append(1)
            out.extend(self.access_selection.to_bytes())
        else:
            out.append(0)

        return out



class LocalDlmsProtocolError(Exception):
    """Protocol error"""


class ApplicationAssociationError(Exception):
    """Something went wrong when trying to setup the application association"""


class PreEstablishedAssociationError(Exception):
    """An error when doing illegal things to the connection if it pre established"""


class ConformanceError(Exception):
    """If APDUs does not match connection Conformance"""


class CipheringError(Exception):
    """Something went wrong when ciphering or deciphering an APDU"""


class DlmsClientException(Exception):
    """An exception that is relating to the client"""


class CommunicationError(Exception):
    """Something went wrong in the communication with a meter"""


class CryptographyError(Exception):
    """Something went wrong then applying a cryptographic function"""


class DecryptionError(CryptographyError):
    """
    Unable to decrypt an APDU. It can be due to mismatch in authentication tag
    because the ciphertext has changed or that the key, nonce or associated data is
    wrong
    """


class NoRlrqRlreError(Exception):
    """
    Is raised from connection when a ReleaseRequest is issued on a connection that has use_rlrq_rlre==False
    Control for client to just skip Release and disconnect the lower layer.
    """


from functools import partial
from typing import *

import attr

from dlms_cosem import a_xdr, dlms_data, security
from dlms_cosem.protocol import xdlms
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.conformance import Conformance

int_from_bytes = partial(int.from_bytes, byteorder="big")


@attr.s(auto_attribs=True)
class InitiateRequest(AbstractXDlmsApdu):
    """
    InitiateRequest ::= SEQUENCE {
    dedicated-key: OCTET STRING OPTIONAL
    response-allowed: BOOLEAN DEFAULT TRUE
    proposed-quality-of-service: IMPLICIT Integer8 OPTIONAL
    proposed-dlms-version-number: Integer8  # Always 6?
    proposed-conformance: Conformance
    client-max-receive-pdu-size: Unsigned16
    }
    """

    TAG: ClassVar[int] = 0x01  # initiateRequest XDLMS-APDU Choice.

    ENCODING_CONF = a_xdr.EncodingConf(
        [
            a_xdr.Attribute(
                attribute_name="dedicated_key",
                create_instance=dlms_data.OctetStringData.from_bytes,
                optional=True,
            ),
            a_xdr.Attribute(
                attribute_name="response_allowed", create_instance=bool, default=True
            ),
            a_xdr.Attribute(
                attribute_name="proposed_quality_of_service",
                create_instance=int_from_bytes,
                length=1,
            ),
            a_xdr.Attribute(
                attribute_name="proposed_dlms_version_number",
                create_instance=int_from_bytes,
                length=1,
            ),
            a_xdr.Attribute(
                attribute_name="rest",
                create_instance=dlms_data.OctetStringData.from_bytes,
                length=9,
            ),
        ]
    )

    proposed_conformance: Conformance
    proposed_quality_of_service: Optional[int] = attr.ib(default=None)
    client_max_receive_pdu_size: int = attr.ib(default=65535)
    proposed_dlms_version_number: int = attr.ib(default=6)
    response_allowed: bool = attr.ib(default=True)
    dedicated_key: Optional[bytes] = attr.ib(default=None)

    @classmethod
    def from_bytes(cls, _bytes: bytes):
        # There is weird decoding here since it is mixed X-ADS and BER....
        data = bytearray(_bytes)
        apdu_tag = data.pop(0)
        if apdu_tag != 0x01:
            raise ValueError(
                f"Data is not a InitiateReques APDU, got apdu tag {apdu_tag}"
            )

        decoder = a_xdr.AXdrDecoder(cls.ENCODING_CONF)
        object_dict = decoder.decode(data)

        # Since the initiate request mixes a-xdr and ber encoding we make some pragmatic
        # one-off handling of that case.

        rest = bytearray(object_dict.pop("rest").value)
        # rest contains ber endoced propesed conformance and max reciec pdu

        conformance_tag = rest[:2]
        if conformance_tag != b"\x5f\x1f":
            raise ValueError(
                f"Didnt receive conformance tag correcly, got {conformance_tag!r}"
            )
        conformance = xdlms.Conformance.from_bytes(data[-5:-2])
        max_pdu_size = int.from_bytes(data[-2:], "big")
        dedicated_key_obj = object_dict.pop("dedicated_key")
        if dedicated_key_obj:
            dedicated_key = bytes(dedicated_key_obj.value)
        else:
            dedicated_key = None
        return cls(
            **object_dict,
            dedicated_key=dedicated_key,
            proposed_conformance=conformance,
            client_max_receive_pdu_size=max_pdu_size,
        )

    def to_bytes(self):
        # Since the initiate request mixes a-xdr and ber encoding we make some pragmatic
        # one-off handling of that case.
        out = bytearray()
        out.append(self.TAG)
        if self.dedicated_key:
            out.append(0x01)
            out.append(len(self.dedicated_key))
            out.extend(self.dedicated_key)
        else:
            out.append(0x00)
        out.append(0x00)
        out.append(0x00)
        out.append(0x06)
        out.extend(b"_\x1f\x04")
        out.extend(self.proposed_conformance.to_bytes())
        out.extend(self.client_max_receive_pdu_size.to_bytes(2, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GlobalCipherInitiateRequest(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 33

    security_control: security.SecurityControlField
    invocation_counter: int
    ciphered_text: bytes

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but got {tag}")

        length = data.pop(0)
        if length != len(data):
            raise ValueError(f"Octetstring is not of correct length")

        security_control = security.SecurityControlField.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        invocation_counter = int.from_bytes(data[:4], "big")
        ciphered_text = bytes(data[4:])

        return cls(security_control, invocation_counter, ciphered_text)

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)

        octet_string_data = bytearray()
        octet_string_data.extend(self.security_control.to_bytes())
        octet_string_data.extend(self.invocation_counter.to_bytes(4, "big"))
        octet_string_data.extend(self.ciphered_text)
        out.append(len(octet_string_data))
        out.extend(octet_string_data)
        return bytes(out)



from dateutil import parser

from dlms_cosem import cosem, enumerations
from dlms_cosem.cosem import selective_access
from dlms_cosem.cosem.selective_access import RangeDescriptor
from dlms_cosem.protocol.xdlms import GetRequestFactory


def test_capture_object_definition():
    x = selective_access.CaptureObject(
        cosem_attribute=cosem.CosemAttribute(
            interface=enumerations.CosemInterface.CLOCK,
            instance=cosem.Obis(0, 0, 1, 0, 0, 255),
            attribute=2,
        ),
        data_index=0,
    )

    assert x.to_bytes() == (
        b"\x02\x04"  # structure of 4 elements
        b"\x12\x00\x08"  # Clock interface class (unsigned long int)
        b"\t\x06\x00\x00\x01\x00\x00\xff"  # Clock instance octet string of 0.0.1.0.0.255
        b"\x0f\x02"  # attribute index  = 2 (integer)
        b"\x12\x00\x00"  # data index  = 0 (long unsigned)
    )


def test_range_descriptor1():
    data = (
        b"\xc0"  # Get request
        b"\x01"  # normal
        b"\xc1"  # invoke id and priority
        b"\x00\x07"  # Profile generic
        b"\x01\x00c\x01\x00\xff"  # 1.0.99.1.0.255
        b"\x02"  # Attribute 2 =  buffer
        b"\x01"  # non default value
        b"\x01"  # descriptor 1 (range-access)
        b"\x02\x04"  # strucutre of 4 elements
        b"\x02\x04"  # strucutre of 4 elements
        b"\x12\x00\x08"  # clock interface class
        b"\t\x06\x00\x00\x01\x00\x00\xff"  # clock instance name. 0.0.1.0.0.255
        b"\x0f\x02"  # attribute 2
        b"\x12\x00\x00"  # data index = 0
        b"\t\x0c\x07\xe2\x06\x01\xff\x00\x03\x00\xff\xff\x88\x80"  # from date
        b"\t\x0c\x07\xe5\x01\x06\xff\x00\x03\x00\xff\xff\xc4\x00"  # to date
        b"\x01\x00"  # all columns
    )
    assert data

    data2 = (
        b"\xc0\x01\xc1\x00\x07\x01\x00c\x01\x00\xff\x02\x01\x01"
        b"\x02\x04"
        b"\x02\x04"
        b"\x12\x00\x08"
        b"\t\x06\x00\x00\x01\x00\x00\xff"
        b"\x0f\x02"
        b"\x12\x00\x00"
        b"\t\x0c\x07\xe2\x02\x0c\xff\x00\x00\x00\x00\x80\x00\x00"
        b"\t\x0c\x07\xe3\x02\x0c\xff\x00\x00\x00\x00\x80\x00\x00"
        b"\x01\x00"
    )
    assert data2


def test_range_descriptor_to_bytes():
    rd = RangeDescriptor(
        restricting_object=selective_access.CaptureObject(
            cosem_attribute=cosem.CosemAttribute(
                interface=enumerations.CosemInterface.CLOCK,
                instance=cosem.Obis(0, 0, 1, 0, 0, 255),
                attribute=2,
            ),
            data_index=0,
        ),
        from_value=parser.parse("2020-01-01T00:03:00+02:00"),
        to_value=parser.parse("2020-01-06T00:03:00+01:00"),
    )
    data = b"\x01\x02\x04\x02\x04\x12\x00\x08\t\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\t\x0c\x07\xe4\x01\x01\xff\x00\x03\x00\x00\xff\x88\x00\t\x0c\x07\xe4\x01\x06\xff\x00\x03\x00\x00\xff\xc4\x00\x01\x00"
    assert rd.to_bytes() == data


def test_parse_range_descriptor():

    """
    Profile: 1 (15 minutes profile)
    From: 01.10.2017 00:00
    To: 01.10.2017 01:00

    """

    data = b"\xc0\x01\xc1\x00\x07\x01\x00c\x01\x00\xff\x02\x01\x01\x02\x04\x02\x04\x12\x00\x08\t\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\t\x0c\x07\xe1\n\x01\x07\x00\x00\x00\x00\xff\xc4\x80\t\x0c\x07\xe1\n\x01\x07\x01\x00\x00\x00\xff\xc4\x80\x01\x00"
    g = GetRequestFactory.from_bytes(data)

    access = (
        b"\x01"  # Optional value used
        b"\x01"  # Access selector
        b"\x02\x04"  # Structure of 4 elements
        b"\x02\x04"  # structure of 4 elements
        b"\x12\x00\x08"  # Clock interface class (unsigned long int)
        b"\t\x06\x00\x00\x01\x00\x00\xff"  # Clock instance octet string of 0.0.1.0.0.255
        b"\x0f\x02"  # attribute index  = 2 (integer)
        b"\x12\x00\x00"  # data index  = 0 (long unsigned)
        b"\t\x0c\x07\xe1\n\x01\x07\x00\x00\x00\x00\xff\xc4\x80"  # from date
        b"\t\x0c\x07\xe1\n\x01\x07\x01\x00\x00\x00\xff\xc4\x80"  # to date
        b"\x01\x00"  # selected_values empty array.
    )
    assert access

    access_selection = g.access_selection

    assert isinstance(access_selection, RangeDescriptor)
    assert access_selection.selected_values is None
    assert (
        access_selection.restricting_object.cosem_attribute.instance.to_string()
        == "0-0:1.0.0.255"
    )



from dlms_cosem import security
from dlms_cosem.connection import XDlmsApduFactory
from dlms_cosem.protocol import xdlms


class TestInitiateRequest:
    def test_parse_simple(self):
        data = bytes.fromhex("01000000065F1F0400007E1F04B0")

        ir = xdlms.InitiateRequest(
            proposed_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                block_transfer_with_action=True,
                multiple_references=True,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            proposed_quality_of_service=0,
            client_max_receive_pdu_size=1200,
            proposed_dlms_version_number=6,
            response_allowed=True,
            dedicated_key=None,
        )

        assert xdlms.InitiateRequest.from_bytes(data) == ir

    def test_encode_simple(self):
        data = bytes.fromhex("01000000065F1F0400007E1F04B0")

        ir = xdlms.InitiateRequest(
            proposed_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                block_transfer_with_action=True,
                multiple_references=True,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            proposed_quality_of_service=0,
            client_max_receive_pdu_size=1200,
            proposed_dlms_version_number=6,
            response_allowed=True,
            dedicated_key=None,
        )

        assert ir.to_bytes() == data

    def test_decode_with_dedicated_key(self):
        data = bytes.fromhex(
            "01011000112233445566778899AABBCCDDEEFF0000065F1F0400007E1F04B0"
        )

        apdu = xdlms.InitiateRequest(
            proposed_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                block_transfer_with_action=True,
                multiple_references=True,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            proposed_quality_of_service=0,
            client_max_receive_pdu_size=1200,
            proposed_dlms_version_number=6,
            response_allowed=True,
            dedicated_key=b'\x00\x11"3DUfw\x88\x99\xaa\xbb\xcc\xdd\xee\xff',
        )

        assert xdlms.InitiateRequest.from_bytes(data) == apdu

    def test_encode_with_dedicated_key(self):
        data = bytes.fromhex(
            "01011000112233445566778899AABBCCDDEEFF0000065F1F0400007E1F04B0"
        )

        apdu = xdlms.InitiateRequest(
            proposed_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                block_transfer_with_action=True,
                multiple_references=True,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            proposed_quality_of_service=0,
            client_max_receive_pdu_size=1200,
            proposed_dlms_version_number=6,
            response_allowed=True,
            dedicated_key=b'\x00\x11"3DUfw\x88\x99\xaa\xbb\xcc\xdd\xee\xff',
        )

        print(data)
        print(apdu.to_bytes())
        assert data == apdu.to_bytes()


class TestInitiateResponse:
    def test_parse_simple(self):
        data = bytes.fromhex("0800065F1F040000501F01F40007")
        ir = xdlms.InitiateResponse(
            negotiated_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=False,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=False,
                block_transfer_with_action=False,
                multiple_references=False,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            server_max_receive_pdu_size=500,
            negotiated_dlms_version_number=6,
            negotiated_quality_of_service=0,
        )
        assert xdlms.InitiateResponse.from_bytes(data) == ir

    def test_encode_simple(self):
        data = bytes.fromhex("0800065F1F040000501F01F40007")
        ir = xdlms.InitiateResponse(
            negotiated_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=False,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=False,
                block_transfer_with_action=False,
                multiple_references=False,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            server_max_receive_pdu_size=500,
            negotiated_dlms_version_number=6,
            negotiated_quality_of_service=0,
        )
        assert data == ir.to_bytes()

    def test_decode_with_non_default_quality(self):
        data = b"\x08\x01\x00\x06_\x1f\x04\x00\x00\x1e\x1d\x04\xc8\x00\x07"
        ir = xdlms.InitiateResponse.from_bytes(data)
        assert ir.negotiated_quality_of_service == 0


class TestGlobalCipherInitiateRequest:
    def test_parse(self):
        data = bytes.fromhex(
            "21303001234567801302FF8A7874133D414CED25B42534D28DB0047720606B175BD52211BE6841DB204D39EE6FDB8E356855"
        )
        apdu = XDlmsApduFactory.apdu_from_bytes(data)

        assert isinstance(apdu, xdlms.GlobalCipherInitiateRequest)

    def test_to_bytes(self):
        data = bytes.fromhex(
            "21303001234567801302FF8A7874133D414CED25B42534D28DB0047720606B175BD52211BE6841DB204D39EE6FDB8E356855"
        )
        apdu = xdlms.GlobalCipherInitiateRequest(
            security_control=security.SecurityControlField(
                security_suite=0,
                authenticated=True,
                encrypted=True,
                broadcast_key=False,
                compressed=False,
            ),
            invocation_counter=19088743,
            ciphered_text=b'\x80\x13\x02\xff\x8axt\x13=AL\xed%\xb4%4\xd2\x8d\xb0\x04w `k\x17[\xd5"\x11\xbehA\xdb M9\xeeo\xdb\x8e5hU',
        )

        assert apdu.to_bytes() == data

    def test_decrypt(self):
        encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
        authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
        system_title = bytes.fromhex("4D4D4D0000BC614E")
        apdu = xdlms.GlobalCipherInitiateRequest(
            security_control=security.SecurityControlField(
                security_suite=0,
                authenticated=True,
                encrypted=True,
                broadcast_key=False,
                compressed=False,
            ),
            invocation_counter=19088743,
            ciphered_text=b'\x80\x13\x02\xff\x8axt\x13=AL\xed%\xb4%4\xd2\x8d\xb0\x04w `k\x17[\xd5"\x11\xbehA\xdb M9\xeeo\xdb\x8e5hU',
        )

        plain_text = security.decrypt(
            security_control=apdu.security_control,
            system_title=system_title,
            invocation_counter=apdu.invocation_counter,
            cipher_text=apdu.ciphered_text,
            key=encryption_key,
            auth_key=authentication_key,
        )

        unciphered_apdu = XDlmsApduFactory.apdu_from_bytes(plain_text)
        print(unciphered_apdu)
        assert isinstance(unciphered_apdu, xdlms.InitiateRequest)

    def test_encrypt(self):
        apdu = xdlms.InitiateRequest(
            proposed_conformance=xdlms.Conformance(
                general_protection=False,
                general_block_transfer=False,
                delta_value_encoding=False,
                attribute_0_supported_with_set=False,
                priority_management_supported=True,
                attribute_0_supported_with_get=True,
                block_transfer_with_get_or_read=True,
                block_transfer_with_set_or_write=True,
                block_transfer_with_action=True,
                multiple_references=True,
                data_notification=False,
                access=False,
                get=True,
                set=True,
                selective_access=True,
                event_notification=True,
                action=True,
            ),
            proposed_quality_of_service=0,
            client_max_receive_pdu_size=1200,
            proposed_dlms_version_number=6,
            response_allowed=True,
            dedicated_key=b'\x00\x11"3DUfw\x88\x99\xaa\xbb\xcc\xdd\xee\xff',
        )

        encryption_key = bytes.fromhex("000102030405060708090A0B0C0D0E0F")
        authentication_key = bytes.fromhex("D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF")
        system_title = bytes.fromhex("4D4D4D0000BC614E")
        security_control = security.SecurityControlField(
            security_suite=0,
            authenticated=True,
            encrypted=True,
            broadcast_key=False,
            compressed=False,
        )
        invocation_counter = 19088743

        ciphered_apdu = xdlms.GlobalCipherInitiateRequest(
            security_control=security.SecurityControlField(
                security_suite=0,
                authenticated=True,
                encrypted=True,
                broadcast_key=False,
                compressed=False,
            ),
            invocation_counter=19088743,
            ciphered_text=b'\x80\x13\x02\xff\x8axt\x13=AL\xed%\xb4%4\xd2\x8d\xb0\x04w `k\x17[\xd5"\x11\xbehA\xdb M9\xeeo\xdb\x8e5hU',
        )

        cipher_text = security.encrypt(
            security_control=security_control,
            system_title=system_title,
            key=encryption_key,
            auth_key=authentication_key,
            invocation_counter=invocation_counter,
            plain_text=apdu.to_bytes(),
        )

        assert cipher_text == ciphered_apdu.ciphered_text



import pytest

from dlms_cosem.protocol.wrappers import DlmsUdpMessage, WrapperHeader

data_examples_encrypted_data_nofication = [
    b"\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00",
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xe6\x03\xd4\xd3W{\x7fd\x994\xe3\xb7\xc7\x19\xa3\xde5\x1a\xb2\x8cz\xc7\xb8\xa1\xe4D\xb8\x96\x91\xe9%\x91\xce\x1e\xb2\x82}\xf97\xa2\xe5@(\x0fb\x11\xf4\x93d\x80/\xa0\xf5\xc4\x13',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xe7\x1c+\xbc?\xfb\x9aN9x\xf2k\xfa\xf5\xe9A\xe2i\xa2\xb6\x1dG\xb46\x1b/[\x1d"\xf5\xa0N\xffp\x8c\x9f\xfbI<@\x16:\x0e\x19x\xb7D\x9c\xec\x9c\xca\xe0\x8d\x19D',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xe8\xb1\xf9[\xdd.\xdbA\xd3V\xdbW\xeeQ, \xc6\xeace:U\xbb\x18q~A\x9fE\xe8\xd3\xb4\xf3C)\xf4\xce\xb2\x1c\x81A\xa7\xe3\xcc\x00\xf0k~-\x98\xd7j\xf4\xb8\x06',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xe9\xfd\x1c&\xa0\xa1\xa8\x8b\x86\xf3\xdc \x10\xb1{\xeb\xa3h\xa3\xb6\xd2\xad\x96SZ\xd4\x1f\x84\xd6\xcbi\xa86]\xb4\x1b\x8c\xac\xb5D\x94v\xc3\xf4 \xe1\x86\xffk\x1b`E\x11p\x08',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xea3\x80\xbdH\x91\x00v\x18]\xa7|\xf9\xd0\xf5v\xc4{\n\xc0\x98\xef\xb3~\xb7u\x89\x8e\x9c\xcde\x02\x13\xa7?&\x9f\x8c{\xea8N\xd3\x88\xe7\xcc\xd2\x05\x06\xfe7;\x06\x8b:',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xeb0J\xf3\x911\xd5\xa6J\x06\xb2\xbb\xa8\xf1\xb9]\xd2+\xfd\xa4]9\xad\xcb\x08\x89\xe3\x03s4\x0f7\xc5\x80\xd3"f\x89>\xc7\'\xae.\xef\xe2\xd1Z8\x89\xab\xd1\x85\x94\x005',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xecho\xf7\xf6\xd0\x9a\x96+\xe5:\xcc\x95\xe1\xe4\xc6\xfeO\xb1[\xfd\xa2\x93\xe2\xae\xcd\x85]\x7f\xaa\xc7\x99\x8cXQ\xce\x038f`E\xa6\xcf\x87\x924V\xf8\xb1+\x02\xb6.\xfc\xed',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xee\xf4\x86`\x0f\xf8\xcf\x8dMA!\xe1B>Q\r\x9c\x87)\xf4\x8b!b\x85t\xfe\x16\xd9\xcbT\x06sL\xefW\x14H\x7f\xf6#\x10\xa4?\x1av\x00L\xa5`\x1b\xbf>\xf9c\x9f',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xefA&8\xb9C\xa0\xfe\xc2,\x8d\x02\xb4\xc4\xb7}\x9es\x8d\x98\xe3q\t\xdb\x85\x12\\\x14\x9f\xa9\xdf=I\xe3\t\xf9\xc3\xa5\xb3\x81\x0b5\xed\x9fVx\xb4\xc7\x81y.\xb8>n+',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xf0N@\xc8{\xde\xb0\xc12\xbfI"\xdf\xc2\x98\xae~pt\xf3\xec_\x1e\x0f\x93\xf36\xfd\x84\xa2\xdf\xb2\xbc\x0b\xed\x80\x84\xf4\xf2\xcf\xebzf\xb1\x16\xd2E\xc8\xb1k\x93\xefM\x1f\x88',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xf1\xce\xef\x1e-\xb6ad\x9a\xbc?\xc4\x1by+\x9a\xd5\xa9\xf0 J\xa16{i\xd5\xdc\x18\x0f\x8c\xd8\xaf\x8d\x99%\x9d\x1d\xfa\x16[\xaa\tg\xb1\xcej\xb9\x8a\xf8\xa5\xdb\x94(\xd3G',
    b'\x00\x01\x00\x01\x00\x01\x00F\xdb\x08/\x19"\x91\x99\x16A\x03;0\x00\x00\x01\xf2\xeb\xae\xa2s\xd5.\xd6V\xc0\x97wM\x08=G%]\x88b\xb57\x1d\xc0l\xf1 \xdcU\x81z;\x91\xc3\x86\xac/g\xca\xf7\x94\x1a=\x01\xb2\xb6|\xdd\x9d{\xbb\x871\x12K',
]


def test_udp_message_from_bytes():
    udp_header_data = b"\x00\x01\x00\x01\x00\x01\x00F"
    dlms_data = b"\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00"

    udp_data = udp_header_data + dlms_data

    message = DlmsUdpMessage.from_bytes(in_data=udp_data)

    assert message.wrapper_header.version == 1
    assert message.wrapper_header.source_wport == 1
    assert message.wrapper_header.destination_wport == 1
    assert message.wrapper_header.version == 1
    assert message.wrapper_header.length == 70
    assert message.data == dlms_data


def test_upd_message_to_bytes():
    udp_header_data = b"\x00\x01\x00\x01\x00\x01\x00F"
    dlms_data = b"\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00"

    udp_data = udp_header_data + dlms_data
    wrapper_header = WrapperHeader(
        source_wport=1, destination_wport=1, length=len(dlms_data)
    )
    message = DlmsUdpMessage(data=dlms_data, wrapper_header=wrapper_header)
    assert message.wrapper_header.version == 1
    assert message.wrapper_header.source_wport == 1
    assert message.wrapper_header.destination_wport == 1
    assert message.wrapper_header.version == 1
    assert message.wrapper_header.length == 70
    assert message.data == dlms_data

    assert message.to_bytes() == udp_data


# def test_udp_parsing():
#     udp = UDPRequest(data_examples_encrypted_data_nofication[0])
#
#     assert (udp) == 2
#
#     # it is a general global ciphering APDU
#
#     a = (b'\x00\x01\x00\x01\x00\x01\x00F'  # UDP wrapper
#          b'\xdb'  # general global ciphering tag
#          b'\x08/\x19"\x91\x99\x16A\x03'  # system title
#          b';0\x00\x00\x01\xf2'  # security Control field = 0b00110000 No compression, unicast, encrypted authenticated. length = 59 bytes = OK
#          b'\xeb\xae\xa2s\xd5.\xd6V\xc0\x97wM\x08=G%]\x88b\xb57\x1d\xc0l\xf1 \xdcU\x81z;\x91\xc3\x86\xac/g\xca\xf7\x94\x1a='
#          b'\x01\xb2\xb6|\xdd\x9d{\xbb\x871\x12K')  # auth tag)



from functools import partial
from typing import *

import attr

import dlms_cosem.utils
from dlms_cosem import a_xdr, cosem, dlms_data
from dlms_cosem import enumerations as enums
from dlms_cosem.cosem import selective_access
from dlms_cosem.dlms_data import (
    VARIABLE_LENGTH,
    AbstractDlmsData,
    DlmsDataFactory,
    decode_variable_integer,
    encode_variable_integer,
)
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.invoke_id_and_priority import InvokeIdAndPriority

get_request_type_from_bytes = partial(enums.GetRequestType.from_bytes, byteorder="big")
get_response_type_from_bytes = partial(
    enums.GetResponseType.from_bytes, byteorder="big"
)


class NullValue:
    def __call__(self):
        return None


def if_falsy_set_none(value):
    if value:
        return value


@attr.s(auto_attribs=True)
class GetRequestNormal(AbstractXDlmsApdu):
    """
    Represents a Get request.

    Get requests are modeled with a choice but we only support the normal one.

    Get requests work in single attributes on interface classes.
    To get a value you would need the interface class, the instance (OBIS) and the
    attribute id.

    Some attributes allow for selective access to the attributes. For example a load
    profile might be read from a specific date or a specific entry.
    """

    TAG: ClassVar[int] = 192
    REQUEST_TYPE: ClassVar[enums.GetRequestType] = enums.GetRequestType.NORMAL

    cosem_attribute: cosem.CosemAttribute = attr.ib(
        validator=attr.validators.instance_of(cosem.CosemAttribute)
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )
    access_selection: Optional[
        Union[selective_access.RangeDescriptor, selective_access.EntryDescriptor]
    ] = attr.ib(default=None, converter=if_falsy_set_none)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.GetRequestType(data.pop(0))
        if type_choice is not enums.GetRequestType.NORMAL:
            raise ValueError(
                "The data for the GetRequest is not for a GetRequestNormal"
            )

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        cosem_attribute_data = data[:9]
        cosem_attribute = cosem.CosemAttribute.from_bytes(cosem_attribute_data)
        data = data[9:]
        has_access_selection = bool(data.pop(0))
        if has_access_selection:
            access_selection = selective_access.AccessDescriptorFactory.from_bytes(data)
        else:
            access_selection = None

        return cls(
            cosem_attribute=cosem_attribute,
            invoke_id_and_priority=invoke_id_and_priority,
            access_selection=access_selection,
        )

    def to_bytes(self):
        # automatically adding the choice for GetRequestNormal.
        out = bytearray()
        out.append(self.TAG)
        out.append(self.REQUEST_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.cosem_attribute.to_bytes())

        if self.access_selection:
            out.extend(b"\x01")
            out.extend(self.access_selection.to_bytes())
        else:
            out.extend(b"\x00")

        return bytes(out)


@attr.s(auto_attribs=True)
class GetRequestNext(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 192
    REQUEST_TYPE: ClassVar[enums.GetRequestType] = enums.GetRequestType.NEXT

    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.GetRequestType(data.pop(0))
        if type_choice is not enums.GetRequestType.NEXT:
            raise ValueError("The data for the GetRequest is not for a GetRequestNext")
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        assert len(data) == 4  # should only be block number left.
        block_number = int.from_bytes(data, "big")
        return cls(block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.REQUEST_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.block_number.to_bytes(4, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GetRequestWithList(AbstractXDlmsApdu):

    TAG: ClassVar[int] = 192
    REQUEST_TYPE: ClassVar[enums.GetRequestType] = enums.GetRequestType.WITH_LIST

    cosem_attributes_with_selection: List[cosem.CosemAttributeWithSelection]
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be {cls.TAG}"
            )

        type_choice = enums.GetRequestType(data.pop(0))
        if type_choice is not enums.GetRequestType.WITH_LIST:
            raise ValueError(
                "The data for the GetRequest is not for a GetRequestWithList"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        number_of_items = data.pop(0)
        cosem_atts = list()
        for i in range(0, number_of_items):
            # Not really happy with the format of this but it works fine.
            c = cosem.CosemAttributeWithSelection.from_bytes(data)
            cosem_atts.append(c)
            data = data[len(c.to_bytes()) :]

        return cls(
            cosem_attributes_with_selection=cosem_atts,
            invoke_id_and_priority=invoke_id_and_priority,
        )

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.REQUEST_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(
            encode_variable_integer(len(self.cosem_attributes_with_selection))
        )  # number of items
        for item in self.cosem_attributes_with_selection:
            out.extend(item.to_bytes())
        return bytes(out)


@attr.s(auto_attribs=True)
class GetRequestFactory:
    """
    The factory will parse the GetRequest and return either a GetRequestNormal,
    GetRequestNext or a GetRequestWithList.
    """

    TAG: ClassVar[int] = 192

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != GetRequestFactory.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be "
                f"{GetRequestFactory.TAG}"
            )
        request_type = enums.GetRequestType(data.pop(0))
        if request_type == enums.GetRequestType.NORMAL:
            return GetRequestNormal.from_bytes(source_bytes)
        elif request_type == enums.GetRequestType.NEXT:
            return GetRequestNext.from_bytes(source_bytes)
        elif request_type == enums.GetRequestType.WITH_LIST:
            return GetRequestWithList.from_bytes(source_bytes)
        else:
            raise ValueError(
                f"Received an enum request type that is not valid for "
                f"GetRequest: {request_type}"
            )


@attr.s(auto_attribs=True)
class GetResponseNormal(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.NORMAL

    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        return cls(bytes(data), invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.append(self.invoke_id_and_priority.to_bytes())
        out.append(0)  # data result choice
        out.extend(self.data)
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseNormalWithError(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.NORMAL

    error: enums.DataAccessResult = attr.ib(
        validator=attr.validators.instance_of(enums.DataAccessResult)
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        choice = data.pop(0)
        if choice != 1:
            raise ValueError(
                f"The data choice is not 1 to indicate error but: {choice}"
            )

        error = enums.DataAccessResult(data.pop(0))

        return cls(error, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # data error choice
        out.extend(self.error.to_bytes(1, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseWithBlock(AbstractXDlmsApdu):
    """
    The data sent in a block response is an OCTET STRING. Not instance of DLMS Data.
    So it has the length encoding first.

    """

    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if last_block:
            raise ValueError(
                f"Last block set to true in a GetResponseWithBlock. Should only be set "
                f"for a GetResponseLastBlock"
            )
        block_number = int.from_bytes(data[:4], "big")

        data = data[4:]
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        data_length, data = dlms_cosem.dlms_data.decode_variable_integer(data)
        if data_length != len(data):
            raise ValueError(
                "The octet string in block data is not of the correct length"
            )

        return cls(bytes(data), block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(0)  # last block == False
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(0)  # data choice = data
        out.extend(
            dlms_cosem.dlms_data.encode_variable_integer(len(self.data))
        )  # octet string length
        out.extend(self.data)

        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseLastBlock(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    data: bytes = attr.ib(validator=attr.validators.instance_of(bytes))
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if not last_block:
            raise ValueError(
                f"Last block is not set to true in a GetResponseLastBlock."
            )
        block_number = int.from_bytes(data[:4], "big")
        data = data[4:]
        choice = data.pop(0)
        if choice != 0:
            raise ValueError(f"The data choice is not 0 to indicate data but: {choice}")

        data_length, data = dlms_cosem.dlms_data.decode_variable_integer(data)
        if data_length != len(data):
            raise ValueError(
                "The octet string in block data is not of the correct length"
            )

        return cls(bytes(data), block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # last block == True
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(0)  # data choice = data
        out.extend(
            dlms_cosem.dlms_data.encode_variable_integer(len(self.data))
        )  # octet string length
        out.extend(self.data)
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseLastBlockWithError(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_BLOCK
    error: enums.DataAccessResult = attr.ib(
        validator=attr.validators.instance_of(enums.DataAccessResult)
    )
    block_number: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag is not correct. Should be {cls.TAG} but is {tag}")
        response_type = enums.GetResponseType(data.pop(0))
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError(
                f"The response type byte: {response_type} is not for a GetResponseNormal"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        last_block = bool(data.pop(0))
        if not last_block:
            raise ValueError(
                f"Last block is not set to true in a GetResponseLastBlock."
            )
        block_number = int.from_bytes(data[:4], "big")
        data = data[4:]
        choice = data.pop(0)
        if choice != 1:
            raise ValueError(
                f"The data choice is not 1 to indicate error but: {choice}"
            )

        assert len(data) == 1
        error = enums.DataAccessResult(data.pop(0))
        return cls(error, block_number, invoke_id_and_priority)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(1)  # last block == True
        out.extend(self.block_number.to_bytes(4, "big"))
        out.append(1)  # data choice = error
        out.extend(self.error.to_bytes(1, "big"))
        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseWithList(AbstractXDlmsApdu):

    TAG: ClassVar[int] = 196
    RESPONSE_TYPE: ClassVar[enums.GetResponseType] = enums.GetResponseType.WITH_LIST

    response_data: List[Union[AbstractDlmsData, enums.DataAccessResult]] = attr.ib(
        factory=list
    )
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        factory=InvokeIdAndPriority,
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    @staticmethod
    def parse_list_response(source_bytes: bytes, amount: int):
        data = bytearray(source_bytes)
        dlms_data_items = list()
        for index in range(0, amount):
            answer_selection = data.pop(0)
            if answer_selection == 0:
                # DLMS data
                parser = dlms_data.DlmsDataParser()
                obj = parser.parse(data, limit=1)
                rest = parser.get_buffer_tail()
                dlms_data_items.append(obj[0])
                data = rest
            elif answer_selection == 1:
                # Data Access Result
                dlms_data_items.append(enums.DataAccessResult(data.pop(0)))
            else:
                raise ValueError("Not a valid answer selection byte")

        return dlms_data_items

    @property
    def result(self) -> List[Any]:
        """
        Converts the response data to python objects if possible
        """
        out = list()
        for item in self.response_data:
            if isinstance(item, enums.DataAccessResult):
                out.append(item)
            else:
                out.append(item.to_python())

        return out

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError("Not a GetResponse APDU")
        response_type = data.pop(0)
        if response_type != cls.RESPONSE_TYPE:
            raise ValueError("Not a GetResponseWithList Apdu")

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        # List of Get-Data-Response.
        list_length = data.pop(0)
        dlms_data = cls.parse_list_response(data, list_length)

        return cls(
            invoke_id_and_priority=invoke_id_and_priority, response_data=dlms_data
        )

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.append(self.RESPONSE_TYPE)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(encode_variable_integer(len(self.response_data)))
        for item in self.response_data:
            if isinstance(item, AbstractDlmsData):
                out.append(0)
                out.extend(item.to_bytes())
            elif isinstance(item, enums.DataAccessResult):
                out.append(1)
                out.append(item.value)

            else:
                raise ValueError(
                    f"unknown data in response for GetResponseWithList: {item}"
                )

        return bytes(out)


@attr.s(auto_attribs=True)
class GetResponseFactory:

    TAG: ClassVar[int] = 196

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != GetResponseFactory.TAG:
            raise ValueError(
                f"Tag is not correct. Should be {GetResponseFactory.TAG} but is {tag}"
            )
        response_type = enums.GetResponseType(data.pop(0))
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        if response_type == enums.GetResponseType.NORMAL:
            # check if it is an error or data response by assesing the choice.
            choice = data.pop(0)
            if choice == 0:
                return GetResponseNormal(
                    invoke_id_and_priority=invoke_id_and_priority, data=bytes(data)
                )
            elif choice == 1:
                assert len(data) == 1  # should only be one byte left.
                error = enums.DataAccessResult(data.pop(0))
                return GetResponseNormalWithError(
                    invoke_id_and_priority=invoke_id_and_priority, error=error
                )
        elif response_type == enums.GetResponseType.WITH_BLOCK:
            last_block = bool(data.pop(0))
            block_number = int.from_bytes(data[:4], "big")
            data = data[4:]
            choice = data.pop(0)
            if choice == 0:
                data_length, data = dlms_cosem.dlms_data.decode_variable_integer(data)
                if data_length != len(data):
                    raise ValueError(
                        "The octet string in block data is not of the correct length"
                    )

                if last_block:
                    return GetResponseLastBlock(
                        bytes(data), block_number, invoke_id_and_priority
                    )
                else:
                    return GetResponseWithBlock(
                        bytes(data), block_number, invoke_id_and_priority
                    )
            elif choice == 1:
                assert len(data) == 1  # should only be one byte left.
                error = enums.DataAccessResult(data.pop(0))
                if last_block:
                    return GetResponseLastBlockWithError(
                        error, block_number, invoke_id_and_priority
                    )
                else:
                    raise ValueError(
                        "It is not possible to send an error on a "
                        "GetResponseWithBlock. When an error occurs it "
                        "should always be sent in a GetResponseLastBlockWithError"
                    )

        elif response_type == enums.GetResponseType.WITH_LIST:
            return GetResponseWithList.from_bytes(bytes(source_bytes))

        else:
            raise ValueError("Response type is not a valid GetResponse type")



import attr


@attr.s(auto_attribs=True)
class WrapperHeader:
    """
    The wrapper header contains 4 parts. Each is an unsigned 16 bit integer.

    * version
    * Source wPort
    * Destination wPort
    * length of the dlms data transferred.

    Reserved wrapper port numbers:

    Client side reserved addresses:

    ==========================      =============
    Description                     wPort number
    =========================       =============
    No-station                      0
    Client Management Process       1
    Public Client                   16
    Open for client SAP assignment  2-15, 17-255

    Server side reserved addresses:

    ==========================      =============
    Description                     wPort number
    =========================       =============
    No-station                      0
    Management Logical Device       1
    Reserved                        2-15
    Open for client SAP assignment  16-126
    All-station (Broad Cast)        127

    :param version: Version of the wrapper. Currently value 0x0001 is used.
    :param source_wport: wPort (Wrapper Port) number for the sending
        DLMS/COSEM Application Entity (AE)
    :param destination_wport: wPort (Wrapper Port) number for the receiving
        DLMS/COSEM Application Entity (AE)
    :param length: Length of data in xDLMS APDU to be transported.

    """

    source_wport: int
    destination_wport: int
    length: int
    version: int = attr.ib(default=1)

    def to_bytes(self):
        _version = self.version.to_bytes(2, "big")
        _source_wport = self.source_wport.to_bytes(2, "big")
        _destination_wport = self.destination_wport.to_bytes(2, "big")
        _length = self.length.to_bytes(2, "big")

        return _version + _source_wport + _destination_wport + _length

    @classmethod
    def from_bytes(cls, in_data):
        if len(in_data) != 8:
            raise ValueError(
                f"Wrapper Header can only consists of 8 bytes and "
                f"got {len(in_data)}"
            )
        version = int.from_bytes(in_data[0:2], "big")
        source_wport = int.from_bytes(in_data[2:4], "big")
        destination_wport = int.from_bytes(in_data[4:6], "big")
        length = int.from_bytes(in_data[6:8], "big")

        return cls(source_wport, destination_wport, length, version)


@attr.s(auto_attribs=True)
class WrapperProtocolDataUnit:
    """

    When sending DLMS data over UDP or TCP you need to include an additional
    wrapper to:

    * Provide additional addressing functionality on top of UDP/TCP port. (Since
      a physical device can host several logical devices)
    * Describe the length of the data sent. Especially for TCP where the data
      can be split up in several packets.

    :param data: The bytes of the xDLMS APDU transported.
    :param WrapperHeader wrapper_header: Wrapper header to declare additional
     information on how to handle the data sent.
    """

    data: bytes
    wrapper_header: WrapperHeader

    def to_bytes(self):
        return self.wrapper_header.to_bytes() + self.data

    @classmethod
    def from_bytes(cls, in_data):
        wrapper_header_data = in_data[0:8]
        data = in_data[8:]

        wrapper_header = WrapperHeader.from_bytes(wrapper_header_data)

        data_length = len(data)
        if not wrapper_header.length == data_length:
            raise ValueError(
                (
                    f"Length of data in Wrapper Protocol Data Unit class "
                    f"{cls.__class__.__name__}, ({data_length}) does not match "
                    f"the length parameter in the Wrapper Header "
                    f"({wrapper_header.length})"
                )
            )

        return cls(data, wrapper_header)


class DlmsUdpMessage(WrapperProtocolDataUnit):
    """
    Handle UPD messages with DLMS APDU content
    """

    pass



from enum import IntEnum
from typing import *

import attr

from dlms_cosem import cosem
from dlms_cosem import enumerations as enums
from dlms_cosem import parsers
from dlms_cosem.cosem.selective_access import CaptureObject, RangeDescriptor


class SortMethod(IntEnum):
    FIFO = 1
    LIFO = 2
    LARGEST = 3
    SMALLEST = 4
    NEAREST_TO_ZERO = 5
    FARTHEST_FROM_ZERO = 6


@attr.s(auto_attribs=True)
class AttributeDescription:
    attribute_id: int
    attribute_name: str
    data_parser: Optional[Any] = attr.ib(default=None)
    data_converter: Optional[Any] = attr.ib(default=None)  # callable?


@attr.s(auto_attribs=True)
class Data:
    INTERFACE_CLASS_ID: ClassVar[enums.CosemInterface] = enums.CosemInterface.DATA

    logical_name: cosem.Obis
    value: Any

    STATIC_ATTRIBUTES: ClassVar[Dict[int, AttributeDescription]] = {
        1: AttributeDescription(attribute_id=1, attribute_name="logical_name"),
    }

    DYNAMIC_ATTRIBUTES: ClassVar[Dict[int, AttributeDescription]] = {
        2: AttributeDescription(attribute_id=2, attribute_name="value"),
    }

    SELECTIVE_ACCESS: ClassVar[Dict[int, Type[RangeDescriptor]]] = {}

    METHODS: ClassVar[Dict[int, str]] = {1: "reset", 2: "capture"}

    DYNAMIC_CONVERTERS: ClassVar[Dict[int, Callable]] = {}

    def is_static_attribute(self, attribute_id: int) -> bool:
        return attribute_id in self.STATIC_ATTRIBUTES.keys()


def convert_load_profile(instance, data):
    parser = parsers.ProfileGenericBufferParser(
        capture_objects=[x.cosem_attribute for x in instance.capture_objects],
        capture_period=instance.capture_period,
    )
    return parser.parse_entries(data)


@attr.s(auto_attribs=True)
class ProfileGeneric:
    INTERFACE_CLASS_ID: ClassVar[
        enums.CosemInterface
    ] = enums.CosemInterface.PROFILE_GENERIC

    logical_name: cosem.Obis
    buffer = List[List[Any]]
    capture_objects: List[CaptureObject]
    capture_period: int
    sort_method: Optional[SortMethod] = attr.ib(default=None)
    sort_object: Optional[CaptureObject] = attr.ib(default=None)
    entries_in_use: Optional[int] = attr.ib(default=None)
    profile_entries: Optional[int] = attr.ib(default=None)

    STATIC_ATTRIBUTES: ClassVar[Dict[int, AttributeDescription]] = {
        1: AttributeDescription(attribute_id=1, attribute_name="logical_name"),
        3: AttributeDescription(attribute_id=3, attribute_name="capture_objects"),
        4: AttributeDescription(attribute_id=4, attribute_name="capture_period"),
        5: AttributeDescription(attribute_id=5, attribute_name="sort_method"),
        6: AttributeDescription(attribute_id=6, attribute_name="sort_object"),
        8: AttributeDescription(attribute_id=8, attribute_name="profile_entries"),
    }

    DYNAMIC_ATTRIBUTES: ClassVar[Dict[int, AttributeDescription]] = {
        2: AttributeDescription(attribute_id=2, attribute_name="buffer"),
        7: AttributeDescription(attribute_id=7, attribute_name="entries_in_use"),
    }

    SELECTIVE_ACCESS: ClassVar[Dict[int, Type[RangeDescriptor]]] = {2: RangeDescriptor}

    METHODS: ClassVar[Dict[int, str]] = {1: "reset", 2: "capture"}

    # Todo: needs to take instance of the class
    DYNAMIC_CONVERTERS: ClassVar[Dict[int, Callable]] = {2: convert_load_profile}

    def reset(self, data: int = 0):
        """clears the buffer"""
        ...

    def capture(self, data: int = 0):
        """initiate a capture"""

    def is_static_attribute(self, attribute_id: int) -> bool:
        return attribute_id in self.STATIC_ATTRIBUTES.keys()



from enum import IntEnum
from typing import *

import attr

from dlms_cosem import cosem, enumerations


class AccessRight(IntEnum):
    READ_ACCESS = 0
    WRITE_ACCESS = 1
    AUTHENTICATED_REQUEST = 2
    ENCRYPTED_REQUEST = 3
    DIGITALLY_SIGNED_REQUEST = 4
    AUTHENTICATED_RESPONSE = 5
    ENCRYPTED_RESPONSE = 6
    DIGITALLY_SIGNED_RESPONSE = 7


@attr.s(auto_attribs=True)
class AttributeAccessRights:
    attribute: int
    access_rights: List[AccessRight]
    access_selectors: List[int] = attr.ib(
        factory=list, converter=attr.converters.default_if_none(factory=list)
    )


@attr.s(auto_attribs=True)
class MethodAccessRights:
    method: int
    access_rights: List[AccessRight]


@attr.s(auto_attribs=True)
class AssociationObjectListItem:
    interface: enumerations.CosemInterface
    logical_name: cosem.Obis
    version: int
    attribute_access_rights: Dict[int, AttributeAccessRights]
    method_access_rights: Dict[int, MethodAccessRights]



from datetime import datetime, timedelta
from typing import *

import attr

from dlms_cosem import a_xdr, cosem, enumerations
from dlms_cosem.cosem import CosemAttribute
from dlms_cosem.cosem.association import (
    AccessRight,
    AssociationObjectListItem,
    AttributeAccessRights,
    MethodAccessRights,
)
from dlms_cosem.time import datetime_from_bytes


@attr.s(auto_attribs=True)
class ColumnValue:

    attribute: CosemAttribute
    value: Any


@attr.s(auto_attribs=True)
class ProfileGenericBufferParser:

    capture_objects: List[CosemAttribute]
    capture_period: int  # minutes

    def parse_bytes(self, profile_bytes: bytes):
        """
        Profile generic are sent as a sequence of A-XDR encoded DlmsData.
        """
        data_decoder = a_xdr.AXdrDecoder(
            encoding_conf=a_xdr.EncodingConf(
                attributes=[a_xdr.Sequence(attribute_name="data")]
            )
        )
        entries: List[List[Any]] = data_decoder.decode(profile_bytes)["data"]

        return self.parse_entries(entries)

    def parse_entries(
        self, entries: List[List[Optional[Any]]]
    ) -> List[List[Optional[ColumnValue]]]:
        """
        Returns a list of columns with the cosem attribut linked to the value with a
        ColumnValue.
        It also sets the timestamp on each column calculated from the prevoius entry
        if the data has been sent compressed using null values
        """
        parsed_entries = list()
        last_entry_timestamp: Optional[datetime] = None
        for entry in entries:
            if len(entry) != len(self.capture_objects):
                raise ValueError(
                    f"Unable to parse ProfileGeneric entry as the amount of columns "
                    f"({len(entry)}) differ from the parsers set capture_object length "
                    f"({len(self.capture_objects)}) "
                )
            parsed_column = list()
            for index, column in enumerate(entry):
                cosem_attribute = self.capture_objects[index]
                if column is not None:
                    if cosem_attribute.interface == enumerations.CosemInterface.CLOCK:
                        # parse as time.
                        value = datetime_from_bytes(column)[
                            0
                        ]  # TODO: do we need clock status?
                        last_entry_timestamp = value
                        parsed_column.append(
                            ColumnValue(attribute=cosem_attribute, value=value)
                        )
                    else:
                        parsed_column.append(
                            ColumnValue(attribute=cosem_attribute, value=column)
                        )
                else:
                    if cosem_attribute.interface == enumerations.CosemInterface.CLOCK:
                        if last_entry_timestamp:
                            value = last_entry_timestamp + timedelta(
                                minutes=self.capture_period
                            )
                            last_entry_timestamp = value
                            parsed_column.append(
                                ColumnValue(attribute=cosem_attribute, value=value)
                            )

                        else:
                            parsed_column.append(None)

            parsed_entries.append(parsed_column)

        return parsed_entries


class AssociationObjectListParser:
    @staticmethod
    def parse_bytes(profile_bytes: bytes):
        """
        Profile generic are sent as a sequence of A-XDR encoded DlmsData.
        """
        data_decoder = a_xdr.AXdrDecoder(
            encoding_conf=a_xdr.EncodingConf(
                attributes=[a_xdr.Sequence(attribute_name="data")]
            )
        )
        entries: List[List[Any]] = data_decoder.decode(profile_bytes)["data"]

        return AssociationObjectListParser.parse_entries(entries)

    @staticmethod
    def parse_access_right(access_right: int) -> List[AccessRight]:
        parsed_access_rights = list()
        if bool(access_right & 0b00000001):
            parsed_access_rights.append(AccessRight.READ_ACCESS.value)
        if bool(access_right & 0b00000010):
            parsed_access_rights.append(AccessRight.WRITE_ACCESS.value)
        if bool(access_right & 0b00000100):
            parsed_access_rights.append(AccessRight.AUTHENTICATED_REQUEST.value)
        if bool(access_right & 0b00001000):
            parsed_access_rights.append(AccessRight.ENCRYPTED_REQUEST.value)
        if bool(access_right & 0b00010000):
            parsed_access_rights.append(AccessRight.DIGITALLY_SIGNED_REQUEST.value)
        if bool(access_right & 0b00100000):
            parsed_access_rights.append(AccessRight.AUTHENTICATED_RESPONSE.value)
        if bool(access_right & 0b01000000):
            parsed_access_rights.append(AccessRight.ENCRYPTED_RESPONSE.value)
        if bool(access_right & 0b10000000):
            parsed_access_rights.append(AccessRight.DIGITALLY_SIGNED_RESPONSE.value)

        return parsed_access_rights

    @staticmethod
    def parse_attribute_access_rights(
        access_rights: List[List[Optional[Union[int, List[int]]]]]
    ):
        parsed_access_rights = list()
        for right in access_rights:
            parsed_access_rights.append(
                AttributeAccessRights(
                    attribute=right[0],
                    access_rights=AssociationObjectListParser.parse_access_right(
                        right[1]
                    ),
                    access_selectors=right[2],
                )
            )
        return parsed_access_rights

    @staticmethod
    def parse_method_access_rights(
        access_rights: List[List[int]],
    ) -> List[MethodAccessRights]:
        parsed_access_rights = list()
        for right in access_rights:
            parsed_access_rights.append(
                MethodAccessRights(
                    method=right[0],
                    access_rights=AssociationObjectListParser.parse_access_right(
                        right[1]
                    ),
                )
            )

        return parsed_access_rights

    @staticmethod
    def parse_entries(object_list):
        parsed_objects = list()
        for obj in object_list:
            interface = enumerations.CosemInterface(obj[0]).value
            version = obj[1]
            logical_name = cosem.Obis.from_bytes(obj[2])
            access_rights = obj[3]
            attribute_access_rights = (
                AssociationObjectListParser.parse_attribute_access_rights(
                    access_rights[0]
                )
            )
            attribute_access_dict = {
                access.attribute: access for access in attribute_access_rights
            }
            method_access_rights = (
                AssociationObjectListParser.parse_method_access_rights(access_rights[1])
            )
            method_access_dict = {
                access.method: access for access in method_access_rights
            }

            parsed_objects.append(
                AssociationObjectListItem(
                    interface=interface,
                    version=version,
                    logical_name=logical_name,
                    attribute_access_rights=attribute_access_dict,
                    method_access_rights=method_access_dict,
                )
            )
        return parsed_objects



import attr
import structlog

from dlms_cosem.exceptions import LocalDlmsProtocolError
from dlms_cosem.protocol import acse, xdlms

LOG = structlog.get_logger()


class _SentinelBase(type):
    """
    Sentinel values

     - Inherit identity-based comparison and hashing from object
     - Have a nice repr
     - Have a *bonus property*: type(sentinel) is sentinel

     The bonus property is useful if you want to take the return value from
     next_event() and do some sort of dispatch based on type(event).

     Taken from h11.
    """

    def __repr__(self):
        return self.__name__


# Some simple flow control classes


@attr.s()
class HlsStart:
    pass


@attr.s()
class HlsSuccess:
    pass


@attr.s()
class HlsFailed:
    pass


@attr.s()
class RejectAssociation:
    pass


@attr.s()
class EndAssociation:
    """
    Is used when settings.use_rlrq_rlre == False to send the state to NO_ASSOCIATION
    """
    pass


def make_sentinel(name):
    cls = _SentinelBase(name, (_SentinelBase,), {})
    cls.__class__ = cls
    return cls


NO_ASSOCIATION = make_sentinel("NO_ASSOCIATION")

AWAITING_ASSOCIATION_RESPONSE = make_sentinel("AWAITING_ASSOCIATION_RESPONSE")

READY = make_sentinel("READY")

AWAITING_RELEASE_RESPONSE = make_sentinel("AWAITING_RELEASE_RESPONSE")
AWAITING_ACTION_RESPONSE = make_sentinel("AWAITING_ACTION_RESPONSE")
AWAITING_GET_RESPONSE = make_sentinel("AWAITING_GET_RESPONSE")
AWAITING_GET_BLOCK_RESPONSE = make_sentinel("AWAITING_GET_BLOCK_RESPONSE")
SHOULD_ACK_LAST_GET_BLOCK = make_sentinel("SHOULD_ACK_LAST_GET_BLOCK")
AWAITING_SET_RESPONSE = make_sentinel("AWAITING_SET_RESPONSE")

SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT = make_sentinel(
    "SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT"
)
AWAITING_HLS_CLIENT_CHALLENGE_RESULT = make_sentinel(
    "AWAITING_HLS_CLIENT_CHALLENGE_RESULT"
)
HLS_DONE = make_sentinel("HLS_DONE")

NEED_DATA = make_sentinel("NEED_DATA")

# TODO: block handling is not working with this state layout.

DLMS_STATE_TRANSITIONS = {
    NO_ASSOCIATION: {acse.ApplicationAssociationRequest: AWAITING_ASSOCIATION_RESPONSE},
    AWAITING_ASSOCIATION_RESPONSE: {
        acse.ApplicationAssociationResponse: READY,
        xdlms.ExceptionResponse: NO_ASSOCIATION,
    },
    READY: {
        acse.ReleaseRequest: AWAITING_RELEASE_RESPONSE,
        xdlms.GetRequestNormal: AWAITING_GET_RESPONSE,
        xdlms.GetRequestWithList: AWAITING_GET_RESPONSE,
        xdlms.SetRequestNormal: AWAITING_SET_RESPONSE,
        HlsStart: SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT,
        RejectAssociation: NO_ASSOCIATION,
        xdlms.ActionRequestNormal: AWAITING_ACTION_RESPONSE,
        xdlms.DataNotification: READY,
        EndAssociation: NO_ASSOCIATION,
    },
    SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT: {
        xdlms.ActionRequestNormal: AWAITING_HLS_CLIENT_CHALLENGE_RESULT
    },
    AWAITING_HLS_CLIENT_CHALLENGE_RESULT: {
        xdlms.ActionResponseNormalWithData: HLS_DONE,
        xdlms.ActionResponseNormal: NO_ASSOCIATION,
        xdlms.ActionResponseNormalWithError: NO_ASSOCIATION,
    },
    HLS_DONE: {HlsSuccess: READY, HlsFailed: NO_ASSOCIATION},
    AWAITING_GET_RESPONSE: {
        xdlms.GetResponseNormal: READY,
        xdlms.GetResponseWithList: READY,
        xdlms.GetResponseWithBlock: SHOULD_ACK_LAST_GET_BLOCK,
        xdlms.GetResponseNormalWithError: READY,
        xdlms.ExceptionResponse: READY,
    },
    AWAITING_GET_BLOCK_RESPONSE: {
        xdlms.GetResponseWithBlock: SHOULD_ACK_LAST_GET_BLOCK,
        xdlms.GetResponseNormalWithError: READY,
        xdlms.ExceptionResponse: READY,
        xdlms.GetResponseLastBlockWithError: READY,
        xdlms.GetResponseLastBlock: READY,
    },
    AWAITING_SET_RESPONSE: {xdlms.SetResponseNormal: READY},
    AWAITING_ACTION_RESPONSE: {
        xdlms.ActionResponseNormal: READY,
        xdlms.ActionResponseNormalWithData: READY,
        xdlms.ActionResponseNormalWithError: READY,
    },
    SHOULD_ACK_LAST_GET_BLOCK: {xdlms.GetRequestNext: AWAITING_GET_BLOCK_RESPONSE},
    AWAITING_RELEASE_RESPONSE: {
        acse.ReleaseResponse: NO_ASSOCIATION,
        xdlms.ExceptionResponse: READY,
    },
}


@attr.s(auto_attribs=True)
class DlmsConnectionState:
    """
    Handles state changes in DLMS, we only focus on Client implementation as of now.

    A DLMS event is passed to `process_event` and it moves the state machine to the
    correct state. If an event is processed that is not set to be able to transition
    the state in the current state a LocalProtocolError is raised.
    """

    current_state: _SentinelBase = attr.ib(default=NO_ASSOCIATION)

    def process_event(self, event):

        self._transition_state(type(event))

    def _transition_state(self, event_type):
        try:
            new_state = DLMS_STATE_TRANSITIONS[self.current_state][event_type]
        except KeyError:
            raise LocalDlmsProtocolError(
                f"can't handle event type {event_type} when state={self.current_state}"
            )
        old_state = self.current_state
        self.current_state = new_state
        LOG.debug(f"DLMS state transitioned", old_state=old_state, new_state=new_state)



from __future__ import annotations  # noqa

import socket
import sys
from typing import Optional, Tuple

from dlms_cosem.hdlc import connection, address, state, frames
from dlms_cosem.protocol.wrappers import WrapperHeader, WrapperProtocolDataUnit

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import attr
import serial

from dlms_cosem import exceptions

from typing import *

import structlog

if TYPE_CHECKING:
    pass

LOG = structlog.get_logger()

LLC_COMMAND_HEADER = b"\xe6\xe6\x00"
LLC_RESPONSE_HEADER = b"\xe6\xe7\x00"


class ClientError(Exception):
    """General error in client"""


class IoImplementation(Protocol):
    def connect(self) -> None: ...

    def disconnect(self) -> None: ...

    def send(self, data: bytes) -> None: ...

    def recv(self, amount: int) -> bytes: ...

    def recv_until(self, end: bytes) -> bytes: ...


class DlmsTransport(Protocol):
    """
    Protocol for a class that should be used for transport.
    """

    client_logical_address: int
    server_logical_address: int
    io: IoImplementation
    timeout: int

    def connect(self) -> None: ...

    def disconnect(self) -> None: ...

    def send_request(self, bytes_to_send: bytes) -> bytes: ...


@attr.s(auto_attribs=True)
class SerialIO:
    port_name: str
    baud_rate: int = attr.ib(default=9600)
    timeout: int = attr.ib(default=10)

    serial_port: Optional[serial.Serial] = attr.ib(init=False, default=None)

    def connect(self):
        if self.serial_port:
            raise RuntimeError(
                f"Trying to open port {self.port_name} when the port "
                f"already is open"
            )
        self.serial_port = serial.Serial(
            port=self.port_name, baudrate=self.baud_rate, timeout=self.timeout
        )

    def disconnect(self):
        if self.serial_port:
            self.serial_port.close()
        self.serial_port = None

    def send(self, data: bytes) -> None:
        if self.serial_port:
            self.serial_port.write(data)
        else:
            raise RuntimeError("Trying to send data on closed serial port")

    def recv(self, amount: int = 1) -> bytes:
        if self.serial_port:
            data = self.serial_port.read(amount)
        else:
            raise RuntimeError("Trying to read data from closed serial port")
        return data

    def recv_until(self, end: bytes) -> bytes:
        if self.serial_port:
            data = self.serial_port.read_until(end)
        else:
            raise RuntimeError("Trying to read data from closed serial port")
        return data


@attr.s(auto_attribs=True)
class BlockingTcpIO:
    """
    A TCP transport using Blocking I/O.
    """

    host: str
    port: int
    timeout: int = attr.ib(default=10)
    tcp_socket: Optional[socket.socket] = attr.ib(init=False, default=None)

    @property
    def address(self) -> Tuple[str, int]:
        return self.host, self.port

    def connect(self):
        """
        Create a new socket and set it on the transport
        """
        if self.tcp_socket:
            raise RuntimeError(f"There is already an active socket to {self.address}")

        try:
            self.tcp_socket = socket.create_connection(
                address=self.address, timeout=self.timeout
            )
        except (
            OSError,
            IOError,
            socket.timeout,
            socket.error,
            ConnectionRefusedError,
        ) as e:
            raise exceptions.CommunicationError("Unable to connect socket") from e
        LOG.info(f"Connected to {self.address}")

    def disconnect(self):
        """
        Close socket and remove it from the transport. No-op if the socket is already
        closed.
        """
        if self.tcp_socket:
            # only disconnect if there is a socket.
            try:
                self.tcp_socket.shutdown(socket.SHUT_RDWR)
                self.tcp_socket.close()
            except (OSError, IOError, socket.timeout, socket.error) as e:
                self.tcp_socket = None
                raise exceptions.CommunicationError from e
            self.tcp_socket = None
            LOG.info(f"Connection to {self.address} is closed")

    def send(self, data: bytes):
        """
        Sends a whole DLMS APDU wrapped in the DLMS IP Wrapper.
        """
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")
        try:
            self.tcp_socket.sendall(data)
        except (OSError, IOError, socket.timeout, socket.error) as e:
            raise exceptions.CommunicationError("Could no send data") from e

    def recv(self, amount: int = 1) -> bytes:
        """
        Receives a whole DLMS APDU. Gets the total length from the DLMS IP Wrapper.
        """
        if not self.tcp_socket:
            raise RuntimeError("TCP transport not connected.")
        data = b""
        while len(data) < amount:
            try:
                data += self.tcp_socket.recv(amount - len(data))
            except (OSError, IOError, socket.timeout, socket.error) as e:
                raise exceptions.CommunicationError("Could not receive data") from e
        return data

    def recv_until(self, end: bytes) -> bytes:
        data = b""
        while not data.endswith(end):
            data += self.recv()
        return data


@attr.s(auto_attribs=True)
class TcpTransport:
    """
    A TCP transport.
    """

    client_logical_address: int
    server_logical_address: int
    io: IoImplementation
    timeout: int = attr.ib(default=10)

    def wrap(self, bytes_to_wrap: bytes) -> bytes:
        """
        When sending data over TCP or UDP it is necessary to wrap the data in the in
        the DLMS IP wrapper. This is so the server (meter) knows where the data is
        intended and how long the message is since there is no "final/poll" bit like
        in HDLC.
        """
        header = WrapperHeader(
            source_wport=self.client_logical_address,
            destination_wport=self.server_logical_address,
            length=len(bytes_to_wrap),
        )
        return WrapperProtocolDataUnit(bytes_to_wrap, header).to_bytes()

    def connect(self):
        self.io.connect()

    def disconnect(self):
        self.io.disconnect()

    def send_request(self, bytes_to_send: bytes) -> bytes:
        """
        Sends a whole DLMS APDU wrapped in the DLMS IP Wrapper.
        """
        wrapped = self.wrap(bytes_to_send)
        LOG.debug("Sending data", data=wrapped, transport=self)
        self.io.send(self.wrap(bytes_to_send))

        return self.recv_response()

    def recv_response(self) -> bytes:
        """
        Receives a whole DLMS APDU. Gets the total length from the DLMS IP Wrapper.
        """
        header_data = self.io.recv(8)
        header = WrapperHeader.from_bytes(header_data)
        data = self.io.recv(header.length)

        LOG.debug("Received data", data=header_data + data, transport=self)

        return data


@attr.s(auto_attribs=True)
class HdlcTransport:
    """
    HDLC transport to send data over serial.
    """

    client_logical_address: int
    server_logical_address: int
    io: IoImplementation
    server_physical_address: Optional[int] = attr.ib(default=None)
    client_physical_address: Optional[int] = attr.ib(default=None)
    extended_addressing: bool = attr.ib(default=False)
    timeout: int = attr.ib(default=10)
    hdlc_connection: connection.HdlcConnection = attr.ib(
        default=attr.Factory(
            lambda self: connection.HdlcConnection(
                self.server_hdlc_address, self.client_hdlc_address
            ),
            takes_self=True,
        )
    )

    _send_buffer: list = attr.ib(factory=list)
    out_buffer: bytearray = attr.ib(init=False, factory=bytearray)
    in_buffer: bytearray = attr.ib(init=False, factory=bytearray)

    @property
    def server_hdlc_address(self):
        return address.HdlcAddress(
            logical_address=self.server_logical_address,
            physical_address=self.server_physical_address,
            address_type="server",
            extended_addressing=self.extended_addressing,
        )

    @property
    def client_hdlc_address(self):
        return address.HdlcAddress(
            logical_address=self.client_logical_address,
            physical_address=self.client_physical_address,
            address_type="client",
            extended_addressing=self.extended_addressing,
        )

    def connect(self):
        """
        Sets up the HDLC Connection by sending a SNRM request.

        """
        self.io.connect()
        # TODO: Implement hdlc parameter negotiation in SNRM frame

        if self.hdlc_connection.state.current_state != state.NOT_CONNECTED:
            raise ClientError(
                f"Client tried to initiate a HDLC connection but connection state was "
                f"not in NOT_CONNECTED but in "
                f"state={self.hdlc_connection.state.current_state}"
            )
        snrm = frames.SetNormalResponseModeFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
        )
        self.out_buffer += self.hdlc_connection.send(snrm)
        self.drain_out_buffer()
        ua_response = self.next_event()
        return ua_response

    def disconnect(self):
        """
        Sends a DisconnectFrame
        :return:
        """
        disc = frames.DisconnectFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
        )

        self.out_buffer += self.hdlc_connection.send(disc)
        self.drain_out_buffer()
        response = self.next_event()
        self.io.disconnect()
        return response

    def next_event(self):
        """
        Will read the serial line until a proper response event is read.
        :return:
        """

        while True:
            # If we already have a complete event buffered internally, just
            # return that. Otherwise, read some data, add it to the internal
            # buffer, and then try again.
            event = self.hdlc_connection.next_event()
            if event is state.NEED_DATA:
                self.hdlc_connection.receive_data(self.recv_frame())
                continue
            return event

    def send_request(self, telegram: bytes) -> bytes:
        """
        Send will make sure the data that needs to be sent i sent.
        The send is the only public function that will return the response data
        when received in full.
        Send will handle fragmentation of data if data is to large to be sent in a
        single HDLC frame.
        :param telegram:
        :return:
        """
        # prepend the LLC
        # The LLC should only be present in the first segmented information frame.
        # So instead we just prepend the data with it we know it will only be in the
        # intial information frame

        self.out_buffer += LLC_COMMAND_HEADER
        self.out_buffer += telegram
        self.drain_out_buffer()
        in_buffer = bytearray()
        while True:
            response = self.next_event()
            in_buffer += response.payload
            if response.segmented and response.final:
                # there is still data but server has send its max window size
                # tell the server to send more.
                rr = frames.ReceiveReadyFrame(
                    destination_address=self.server_hdlc_address,
                    source_address=self.client_hdlc_address,
                    receive_sequence_number=self.hdlc_connection.server_rsn,
                )
                self.out_buffer += self.hdlc_connection.send(rr)
                self.drain_out_buffer()

            if response.segmented and not response.final:
                # the server will send more frames.
                continue
            if not response.segmented and response.final:
                # this was the last frame
                break

        if not in_buffer.startswith(LLC_RESPONSE_HEADER):
            raise ValueError("The data is not prepended by the LLC response header")
        # don't return the LLC
        return in_buffer[3:]

    def drain_out_buffer(self):
        """
        If the data we need to send is longer than the allowed InformationFrame payload
        size we need to segment the data. It is done by splitting the payload into
        several frames and setting the segmented flag.
        To indicate that we are done with the sending a window we set final on the
        last I-frame. To indicated we are done sending all the data we set segmented to
        False
        :return:
        """
        data_size = self.hdlc_connection.max_data_size
        while len(self.out_buffer) > 0:
            data = self.out_buffer[:data_size]
            self.out_buffer = self.out_buffer[data_size:]
            segmented = bool(self.out_buffer)
            if self.hdlc_connection.state.current_state != state.IDLE:
                LOG.debug("Sending data", data=data, transport=self)
                self.io.send(data)
                return
            # We don't handle window sizes so final is always true
            out_frame = self.generate_information_frame(
                data, segmented=segmented, final=True
            )
            self.send_frame(out_frame)
            # if it is the last frame we should not listen to possible RR frame
            if segmented:
                response = self.next_event()
                if isinstance(response, frames.ReceiveReadyFrame):
                    # send the next information frame.
                    continue
            break

    def generate_information_frame(
        self, payload: bytes, segmented: bool, final: bool
    ) -> frames.InformationFrame:
        return frames.InformationFrame(
            destination_address=self.server_hdlc_address,
            source_address=self.client_hdlc_address,
            payload=payload,
            send_sequence_number=self.hdlc_connection.server_ssn,
            receive_sequence_number=self.hdlc_connection.server_rsn,
            segmented=segmented,
            final=final,
        )

    def send_frame(self, frame):
        LOG.info(f"Sending HDLC frame", frame=frame)
        frame_bytes = self.hdlc_connection.send(frame)
        LOG.debug("Sending data", data=frame_bytes, transport=self)
        self.io.send(frame_bytes)

    def recv_frame(self) -> bytes:
        in_bytes = self.io.recv_until(end=frames.HDLC_FLAG)

        if in_bytes == frames.HDLC_FLAG:
            # We found the first HDLC Frame Flag. We should read until the last one.
            in_bytes += self.io.recv_until(frames.HDLC_FLAG)

        LOG.debug("Received data", data=in_bytes, transport=self)

        return in_bytes

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()



from dlms_cosem.connection import XDlmsApduFactory
from dlms_cosem.protocol.xdlms import DataNotification, GeneralGlobalCipher
from dlms_cosem.security import SecurityControlField


def test_gen_glo_cipher_load():
    dlms_data = b"\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00"

    system_title = b'/\x19"\x91\x99\x16A\x03'

    apdu = XDlmsApduFactory.apdu_from_bytes(apdu_bytes=dlms_data)

    assert isinstance(apdu, GeneralGlobalCipher)

    assert apdu.system_title == system_title


def test_gen_glo_cipher_to_apdu():
    dlms_data = b"\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00"

    system_title = b'/\x19"\x91\x99\x16A\x03'

    apdu = XDlmsApduFactory.apdu_from_bytes(apdu_bytes=dlms_data)

    assert isinstance(apdu, GeneralGlobalCipher)

    assert apdu.system_title == system_title

    unportected_apdu_data = apdu.to_plain_apdu(
        encryption_key=b"MYDUMMYGLOBALKEY", authentication_key=b"MYDUMMYGLOBALKEY"
    )
    unportected_apdu = XDlmsApduFactory.apdu_from_bytes(
        apdu_bytes=unportected_apdu_data
    )

    assert isinstance(unportected_apdu, DataNotification)


def test_gen_glo_cipher_to_bytes():
    dlms_data = b"\xdb\x08/\x19\"\x91\x99\x16A\x03;0\x00\x00\x01\xe5\x02\\\xe9\xd2'\x1f\xd7\x8b\xe8\xc2\x04!\x1a\x91j\x9d\x7fX~\nz\x81L\xad\xea\x89\xe9Y?\x01\xf9.\xa8\xc0\x87\xb5\xbd\xfd\xef\xea\xb6\xbe\xcf(-\xfeI\xc0\x8f[\xe6\xdc\x84\x00"

    apdu = XDlmsApduFactory.apdu_from_bytes(apdu_bytes=dlms_data)
    assert apdu.to_bytes() == dlms_data


def test_data_notification_apdu():
    dlms_data = b'\x0f\x00\x00\x01\xdb\x00\t"\x12Z\x85\x916\x00\x00\x00\x00I\x00\x00\x00\x11\x00\x00\x00\nZ\x85\x13\xd0\x14\x80\x00\x00\x00\r\x00\x00\x00\n\x01\x00'

    apdu = XDlmsApduFactory.apdu_from_bytes(apdu_bytes=dlms_data)

    assert isinstance(apdu, DataNotification)

    print(apdu)


def test_gen_glo_cipher_with_no_system_title_encodes_correct():
    ciphered = GeneralGlobalCipher(
        system_title=None,
        security_control=SecurityControlField(
            security_suite=0, authenticated=True,
            encrypted=True, broadcast_key=False,
            compressed=False),
        invocation_counter=2147483857,
        ciphered_text=b'\x81\xec\x9e\xc4\xbfS\xe9wn\xf0\xc4S\x96\x9f\xbd\xfe\x11\xbe\x9by\x1a\xac\xc0\xff\x8c')

    correct_result = b"\xdb\x00\x1e0\x80\x00\x00\xd1\x81\xec\x9e\xc4\xbfS\xe9wn\xf0\xc4S\x96\x9f\xbd\xfe\x11\xbe\x9by\x1a\xac\xc0\xff\x8c"
    assert ciphered.to_bytes() == correct_result



from typing import *

import attr

from dlms_cosem import cosem, enumerations
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.protocol.xdlms.invoke_id_and_priority import InvokeIdAndPriority

# TODO:  Use same kind of setup as with GET.
# Several classes depending on the type of Action Request/Response
# ActionRequestNormal, ActionResponseNormal, ActionResponseNormalWithError,


@attr.s(auto_attribs=True)
class ActionRequestNormal(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 195
    ACTION_TYPE: ClassVar[enumerations.ActionType] = enumerations.ActionType.NORMAL

    cosem_method: cosem.CosemMethod = attr.ib(
        validator=attr.validators.instance_of(cosem.CosemMethod)
    )
    data: Optional[bytes] = attr.ib(default=None)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        default=InvokeIdAndPriority(0, True, True),
        validator=attr.validators.instance_of(InvokeIdAndPriority),
    )

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)
        out.append(self.ACTION_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.extend(self.cosem_method.to_bytes())
        if self.data:
            out.append(0x01)
            out.extend(self.data)
        else:
            out.append(0x00)
        return bytes(out)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag {tag} is not the correct tag for an ActionRequest, should "
                f"be {cls.TAG}"
            )
        request_type = enumerations.ActionType(data.pop(0))

        if request_type != enumerations.ActionType.NORMAL:
            raise ValueError(
                f"Bytes are not representing a ActionRequestNormal. Action type "
                f"is {request_type}"
            )
        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )
        cosem_method = cosem.CosemMethod.from_bytes(data[:9])
        has_data = bool(data[9])
        if has_data:
            request_data = data[10:]
        else:
            request_data = None

        return cls(
            cosem_method=cosem_method,
            data=request_data,
            invoke_id_and_priority=invoke_id_and_priority,
        )


@attr.s(auto_attribs=True)
class ActionRequestFactory:
    """
    Factory that will parse the ActionRequest and return the correct class for the
    particular instance
    """

    TAG: ClassVar[int] = 195

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != ActionRequestFactory.TAG:
            raise ValueError(
                f"Tag for GET request is not correct. Got {tag}, should be "
                f"{ActionRequestFactory.TAG}"
            )
        request_type = enumerations.ActionType(data.pop(0))
        if request_type == enumerations.ActionType.NORMAL:
            return ActionRequestNormal.from_bytes(source_bytes)
        else:
            raise NotImplementedError(
                f"no class to support action request type {request_type}"
            )


@attr.s(auto_attribs=True)
class ActionResponseNormal(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 199
    ACTION_TYPE: ClassVar[enumerations.ActionType] = enumerations.ActionType.NORMAL

    status: enumerations.ActionResultStatus
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        default=InvokeIdAndPriority(0, True, True)
    )

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)
        out.append(self.ACTION_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(self.status.value)
        out.extend(b"\x00")

        return bytes(out)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag {tag} is not correct for ActionResponse. Should be {cls.TAG}"
            )
        action_type = enumerations.ActionType(data.pop(0))

        if action_type != enumerations.ActionType.NORMAL:
            raise ValueError(
                f"Bytes are not representing a ActionResponseNormal. Action type "
                f"is {action_type}"
            )

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        status = enumerations.ActionResultStatus(data.pop(0))
        has_data = bool(data.pop(0))
        if has_data:
            raise ValueError(
                f"ActionResponse has data and should not be a " f"ActionResponseNormal"
            )

        return cls(invoke_id_and_priority=invoke_id_and_priority, status=status)


@attr.s(auto_attribs=True)
class ActionResponseNormalWithData(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 199
    ACTION_TYPE: ClassVar[enumerations.ActionType] = enumerations.ActionType.NORMAL

    status: enumerations.ActionResultStatus
    data: bytes = attr.ib(default=None)
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        default=InvokeIdAndPriority(0, True, True)
    )

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)
        out.append(self.ACTION_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(self.status.value)
        out.extend(b"\x01")  # has data
        out.extend(b"\x00")  # data result choice
        out.extend(self.data)
        return bytes(out)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag {tag} is not correct for ActionResponse. Should be {cls.TAG}"
            )
        action_type = enumerations.ActionType(data.pop(0))

        if action_type != enumerations.ActionType.NORMAL:
            raise ValueError(
                f"Bytes are not representing a ActionResponseNormal. Action type "
                f"is {action_type}"
            )

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        status = enumerations.ActionResultStatus(data.pop(0))
        has_data = bool(data.pop(0))
        if has_data:
            data_is_result = data.pop(0) == 0
            if not data_is_result:
                raise ValueError(
                    "Data is not a ActionResponseNormalWithData, maybe a "
                    "ActionResponseNormalWithError"
                )
            response_data = data

        else:
            raise ValueError(
                f"ActionResponseNormalWithData does not contain any data. "
                f"Should probably be an ActionResponseNormal"
            )

        return cls(
            invoke_id_and_priority=invoke_id_and_priority,
            status=status,
            data=response_data,
        )


@attr.s(auto_attribs=True)
class ActionResponseNormalWithError(AbstractXDlmsApdu):
    TAG: ClassVar[int] = 199
    ACTION_TYPE: ClassVar[enumerations.ActionType] = enumerations.ActionType.NORMAL

    status: enumerations.ActionResultStatus
    error: enumerations.DataAccessResult
    invoke_id_and_priority: InvokeIdAndPriority = attr.ib(
        default=InvokeIdAndPriority(0, True, True)
    )

    def to_bytes(self):
        out = bytearray()
        out.append(self.TAG)
        out.append(self.ACTION_TYPE.value)
        out.extend(self.invoke_id_and_priority.to_bytes())
        out.append(self.status.value)

        out.extend(b"\x01")
        out.extend(b"\x01")  # data result data (error) choice
        out.append(self.error.value)

        return bytes(out)

    @classmethod
    def from_bytes(cls, source_bytes: bytes):

        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(
                f"Tag {tag} is not correct for ActionResponse. Should be {cls.TAG}"
            )
        action_type = enumerations.ActionType(data.pop(0))

        if action_type != enumerations.ActionType.NORMAL:
            raise ValueError(
                f"Bytes are not representing a ActionResponseNormal. Action type "
                f"is {action_type}"
            )

        invoke_id_and_priority = InvokeIdAndPriority.from_bytes(
            data.pop(0).to_bytes(1, "big")
        )

        status = enumerations.ActionResultStatus(data.pop(0))
        has_data = bool(data.pop(0))
        if has_data:
            data_is_error = data.pop(0) == 1
            if not data_is_error:
                raise ValueError(
                    "Data is not a ActionResponseNormalWithError, maybe a "
                    "ActionResponseNormal"
                )
            assert len(data) == 1
            error = enumerations.DataAccessResult(data.pop(0))

        else:
            raise ValueError("No error data in ActionResponseWithError")

        return cls(
            invoke_id_and_priority=invoke_id_and_priority, status=status, error=error
        )


@attr.s(auto_attribs=True)
class ActionResponseFactory:

    """
    Action-Response ::= CHOICE
    {
    action-response-normal      [1] IMPLICIT Action-Response-Normal
    action-response-with-pblock [2] IMPLICIT Action-Response-With-Pblock,
    action-response-with-list   [3] IMPLICIT Action-Response-With-List,
    action-response-next-pblock [4] IMPLICIT Action-Response-Next-Pblock,
    }

    Action-Response-Normal ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    single-response         Action-Response-With-Optional-Data
    }

    Action-Response-With-Pblock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    pblock                  DataBlock-SA
    }

    Action-Response-With-List ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    list-of-responses       SEQUENCE OF Action-Response-With-Optional-Data
    }

    Action-Response-Next-Pblock ::= SEQUENCE
    {
    invoke-id-and-priority  Invoke-Id-And-Priority,
    block-number            Unsigned32
    }

    Action-Response-With-Optional-Data ::= SEQUENCE
    {
    result              Action-Result,
    return-parameters   Get-Data-Result OPTIONAL
    }

    Get-Data-Result ::= CHOICE
    {
    data                [0] Data,
    ata-access-result   [1] IMPLICIT Data-Access-Result
    }

    """

    TAG: ClassVar[int] = 199

    @staticmethod
    def from_bytes(source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != ActionResponseFactory.TAG:
            raise ValueError(
                f"Tag is not correct. Should be {ActionResponseFactory.TAG} but is {tag}"
            )
        response_type = enumerations.ActionType(data.pop(0))

        data.pop(0)  # Invoke id and priority that is not needed for parsing

        if response_type == enumerations.ActionType.NORMAL:
            data.pop(0)  # Action result status, not needed for parsing
            # check if it is an error or data response by assesing the choice.
            has_data = bool(data.pop(0))
            if has_data:
                choice = data.pop(0)
                if choice == 0:
                    return ActionResponseNormalWithData.from_bytes(source_bytes)
                elif choice == 1:
                    return ActionResponseNormalWithError.from_bytes(source_bytes)
            else:
                return ActionResponseNormal.from_bytes(source_bytes)
        else:
            raise NotImplementedError(
                "Only implemented the ActionResponse Normal "
                "class types is not implemented."
            )



import abc
import datetime
from typing import *
from typing import List, Optional

import attr

from dlms_cosem import time

VARIABLE_LENGTH = -1


class AbstractDlmsData(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, source_bytes: bytes):
        pass

    @abc.abstractmethod
    def to_python(self) -> Any:
        pass

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        pass


@attr.s(auto_attribs=True)
class BaseDlmsData(AbstractDlmsData):
    TAG: ClassVar[int] = 0
    LENGTH: ClassVar[int] = 0

    value: Any

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        raise NotImplementedError(
            f"{cls.__name__} have not implemented byte conversion"
        )

    def to_python(self) -> Any:
        return self.value

    def value_to_bytes(self) -> bytes:
        raise NotImplementedError("value_to_bytes must be implemented in subclass")

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        value_bytes = self.value_to_bytes()
        if self.LENGTH == VARIABLE_LENGTH:
            out.append(len(value_bytes))
        out.extend(value_bytes)
        return bytes(out)


@attr.s(auto_attribs=True)
class NullData(BaseDlmsData):
    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(None)

    def to_python(self) -> Any:
        return None

    TAG = 0


@attr.s(auto_attribs=True)
class DataArray(BaseDlmsData):
    """Sequence of Data"""

    TAG = 1
    LENGTH = VARIABLE_LENGTH

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.extend(encode_variable_integer(len(self.value)))
        for item in self.value:
            out.extend(item.to_bytes())
        return bytes(out)

    def to_python(self) -> List[Any]:
        values = list()
        for item in self.value:
            values.append(item.to_python())
        return values


@attr.s(auto_attribs=True)
class DataStructure(BaseDlmsData):
    """SEQUENCE of Data"""

    TAG = 2
    LENGTH = VARIABLE_LENGTH

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        out.extend(encode_variable_integer(len(self.value)))
        for item in self.value:
            out.extend(item.to_bytes())
        return bytes(out)

    def to_python(self) -> List[Any]:
        values = list()
        for item in self.value:
            values.append(item.to_python())
        return values


@attr.s(auto_attribs=True)
class BooleanData(BaseDlmsData):
    TAG = 3
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        value = bool(int.from_bytes(bytes_data, "big"))
        return cls(value)  # TODO: test this.


@attr.s(auto_attribs=True)
class BitStringData(BaseDlmsData):
    TAG = 4
    LENGTH = VARIABLE_LENGTH

    def value_to_bytes(self) -> bytes:
        return self.value


@attr.s(auto_attribs=True)
class DoubleLongData(BaseDlmsData):
    """32 bit integer"""

    TAG = 5
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(4, "big", signed=True)


@attr.s(auto_attribs=True)
class DoubleLongUnsignedData(BaseDlmsData):
    """32 bit unsigned integer"""

    TAG = 6
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(4, "big")


@attr.s(auto_attribs=True)
class OctetStringData(BaseDlmsData):
    TAG = 9
    LENGTH = VARIABLE_LENGTH

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=bytes_data)

    def value_to_bytes(self) -> bytes:
        return self.value

    def to_python(self) -> bytes:
        return self.value


@attr.s(auto_attribs=True)
class VisibleStringData(BaseDlmsData):
    TAG = 10
    LENGTH = VARIABLE_LENGTH

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=bytes_data.decode("ascii"))

    def value_to_bytes(self) -> bytes:
        return self.value.encode("ascii")
        

@attr.s(auto_attribs=True)
class UTF8StringData(BaseDlmsData):
    TAG = 12
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class BCDData(BaseDlmsData):
    TAG = 13
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class IntegerData(BaseDlmsData):
    """"8 bit integer"""

    TAG = 15
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(1, "big", signed=True)


@attr.s(auto_attribs=True)
class LongData(BaseDlmsData):
    """16  bit integer"""

    TAG = 16
    LENGTH = 2

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(2, "big", signed=True)


@attr.s(auto_attribs=True)
class UnsignedIntegerData(BaseDlmsData):
    """8 bit unsigned integer"""

    TAG = 17
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class UnsignedLongData(BaseDlmsData):
    """16 bit unsigned integer"""

    TAG = 18
    LENGTH = 2

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(2, "big")


@attr.s(auto_attribs=True)
class CompactArrayData(BaseDlmsData):
    """
    Contains a Type description and arrray content in form of octet string
    content_description -> Type Description tag = 0
    array_content -> Octet string  tag = 1
    """

    TAG = 19
    LENGTH = VARIABLE_LENGTH


@attr.s(auto_attribs=True)
class Long64Data(BaseDlmsData):
    """
    64 bit integer
    """

    TAG = 20
    LENGTH = 8

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big", signed=True))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(8, "big", signed=True)


@attr.s(auto_attribs=True)
class UnsignedLong64Data(BaseDlmsData):
    """
    64 bit unsigned integer
    """

    TAG = 21
    LENGTH = 8

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(8, "big")


@attr.s(auto_attribs=True)
class EnumData(BaseDlmsData):
    """
    8 bit integer
    """

    TAG = 22
    LENGTH = 1

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        return cls(value=int.from_bytes(bytes_data, "big"))

    def value_to_bytes(self) -> bytes:
        return self.value.to_bytes(1, "big")


@attr.s(auto_attribs=True)
class Float32Data(BaseDlmsData):
    """
    Octet string of 4 bytes
    """

    TAG = 23
    LENGTH = 4


@attr.s(auto_attribs=True)
class Float64Data(BaseDlmsData):
    """
    Octet string of 8 bytes
    """

    TAG = 24
    LENGTH = 8


@attr.s(auto_attribs=True)
class DateTimeData(BaseDlmsData):
    """Octet string of 12 bytes"""

    TAG = 25
    LENGTH = 12

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Datetime should be 12 bytes long, got {len(bytes_data)}")
        return cls(time.datetime_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class DateData(BaseDlmsData):
    """Octet string of 5 bytes"""

    TAG = 26
    LENGTH = 5

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Date should be 5 bytes long, got {len(bytes_data)}")
        return cls(time.date_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class TimeData(BaseDlmsData):
    """Octet string of 4 bytes"""

    TAG = 27
    LENGTH = 4

    @classmethod
    def from_bytes(cls, bytes_data: bytes):
        if len(bytes_data) != cls.LENGTH:
            raise ValueError(f"Time should be 4 bytes long, got {len(bytes_data)}")
        return cls(time.time_from_bytes(bytes_data))


@attr.s(auto_attribs=True)
class DontCareData(BaseDlmsData):
    """Nulldata"""

    TAG = 255
    LENGTH = 0


@attr.s(auto_attribs=True)
class UnixTimestamp(DoubleLongUnsignedData):
    """Unix timestamps should be represented as double long unsigned"""

    DEFAULT_TIMEZONE = datetime.timezone.utc

    @classmethod
    def from_bytes(cls, bytes_data):
        val = datetime.datetime.fromtimestamp(
            int.from_bytes(bytes_data, "big"), cls.DEFAULT_TIMEZONE
        )
        return cls(value=val)


@attr.s(auto_attribs=True)
class DlmsDataFactory:
    MAP: ClassVar[Dict[int, Type]] = {
        0: NullData,
        1: DataArray,
        2: DataStructure,
        3: BooleanData,
        4: BitStringData,
        5: DoubleLongData,
        6: DoubleLongUnsignedData,
        9: OctetStringData,
        10: VisibleStringData,
        12: UTF8StringData,
        13: BCDData,
        15: IntegerData,
        16: LongData,
        17: UnsignedIntegerData,
        18: UnsignedLongData,
        19: CompactArrayData,
        20: Long64Data,
        21: UnsignedLong64Data,
        22: EnumData,
        23: Float32Data,
        24: Float64Data,
        25: DateTimeData,
        26: DateData,
        27: TimeData,
        255: DontCareData,
    }

    @classmethod
    def get_data_class(cls, tag: int):
        return cls.MAP[tag]


@attr.s(auto_attribs=True)
class DlmsDataParser:

    buffer: bytearray = attr.ib(factory=bytearray, init=False)
    pointer: int = attr.ib(default=0, init=False)
    data: List[AbstractDlmsData] = attr.ib(factory=list, init=False)

    @property
    def buffer_empty(self) -> bool:
        return self.pointer == len(self.buffer)

    def parse(self, data: bytes, limit: Optional[int] = None):
        # clear previous results
        self.data = list()
        self.buffer = bytearray()
        self.pointer = 0
        # fill the buffer
        self.buffer += data

        while not self.buffer_empty:
            self.data.append(self.parse_one_entry())
            if limit:
                if len(self.data) >= limit:
                    break

        return self.data

    def parse_one_entry(self):

        tag = self.get_bytes(1)
        klass = DlmsDataFactory.get_data_class(int.from_bytes(tag, "big"))
        if klass == DataArray:
            return self.decode_array()
        elif klass == DataStructure:
            return self.decode_structure()
        else:

            return self.decode_data(klass)

    def get_buffer_tail(self) -> bytearray:
        return self.buffer[self.pointer :]

    def decode_data(self, data_class) -> AbstractDlmsData:
        if data_class.LENGTH == VARIABLE_LENGTH:
            length = self.decode_variable_integer()
            return data_class.from_bytes(self.get_bytes(length))
        else:
            return data_class.from_bytes(self.get_bytes(data_class.LENGTH))

    def decode_array(self) -> DataArray:
        item_count = self.decode_variable_integer()
        elements = list()
        for _ in range(0, item_count):
            elements.append(self.parse_one_entry())
        return DataArray(value=elements)

    def decode_structure(self) -> DataStructure:
        item_count = self.decode_variable_integer()
        elements = list()
        for _ in range(0, item_count):
            elements.append(self.parse_one_entry())

        return DataStructure(value=elements)

    def get_bytes(self, length: int) -> bytearray:
        """Gets some bytes from the buffer and moves the pointer forward."""
        part = self.buffer[self.pointer : self.pointer + length]
        self.pointer += length
        return part

    @property
    def remaining_buffer(self) -> bytearray:
        return self.buffer[self.pointer :]

    def decode_variable_integer(self) -> int:
        length_data = bytearray()
        first_byte = int.from_bytes(self.get_bytes(1), "big")
        length_is_multiple_bytes = bool(first_byte & 0b10000000)
        if not length_is_multiple_bytes:
            return first_byte
        number_of_bytes_representing_the_length = first_byte & 0b01111111
        for _ in range(0, number_of_bytes_representing_the_length):
            length_data.extend(self.get_bytes(1))
        return int.from_bytes(length_data, "big")


def decode_variable_integer(bytes_input: bytes):
    """
    If the length is fitting in 7 bits it can be encoded in 1 bytes.
    If it is larger then 7 bybitstes the last bit of the first byte indicates
    that the length of the lenght is encoded in the first byte and the length
    is encoded in the following bytes.
    Ex. 0b00000010 -> Length = 2
    Ex 0b100000010, 0b000001111, 0b11111111 -> Lenght = 4095
    :param bytes_input: Input where the variable integer is at the beginning of
    the bytes
    :return: First variable integer the function finds. and the residual bytes
    """

    # is the length encoded in single byte or mutliple?
    is_mutliple_bytes = bool(bytes_input[0] & 0b10000000)
    if is_mutliple_bytes:
        length_length = int(bytes_input[0] & 0b01111111)
        length_data = bytes_input[1 : (length_length + 1)]
        length = int.from_bytes(length_data, "big")
        return length, bytes_input[length_length + 1 :]

    else:
        length = int(bytes_input[0] & 0b01111111)
        return length, bytes_input[1:]


def encode_variable_integer(length: int):
    if length > 0b01111111:
        encoded_length = 1
        while True:
            try:
                length.to_bytes(encoded_length, "big")
            except OverflowError:
                encoded_length += 1
                continue
            break

        length_byte = (0b10000000 + encoded_length).to_bytes(1, "big")
        return length_byte + length.to_bytes(encoded_length, "big")

    else:
        return length.to_bytes(1, "big")



import pytest

from dlms_cosem import enumerations
from dlms_cosem.protocol import acse

# Example encodings from DLMS Green book v10 page:
from dlms_cosem.protocol.acse import UserInformation
from dlms_cosem.protocol.xdlms import Conformance, InitiateResponse
from dlms_cosem.protocol.xdlms.confirmed_service_error import ConfirmedServiceError


class TestDecodeAARE:
    def test_no_ciphering_no_security_ok_association(self):
        data = bytes.fromhex(
            "6129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F040000501F01F40007"
        )
        aare = acse.ApplicationAssociationResponse.from_bytes(data)
        assert not aare.ciphered
        assert aare.result == enumerations.AssociationResult.ACCEPTED
        assert (
            aare.result_source_diagnostics
            == enumerations.AcseServiceUserDiagnostics.NULL
        )
        assert aare.user_information is not None

    def test_no_ciphering_no_security_app_context_name_wrong(self):
        data = bytes.fromhex(
            "6129A109060760857405080101A203020101A305A103020102BE10040E0800065F1F040000501F01F40007"
        )
        aare = acse.ApplicationAssociationResponse.from_bytes(data)
        assert not aare.ciphered
        assert aare.result == enumerations.AssociationResult.REJECTED_PERMANENT
        assert (
            aare.result_source_diagnostics
            == enumerations.AcseServiceUserDiagnostics.APPLICATION_CONTEXT_NAME_NOT_SUPPORTED
        )
        assert isinstance(aare.user_information.content, InitiateResponse)

    def test_no_cipher_no_security_incorrect_dlms_version(self):
        data = bytes.fromhex(
            "611FA109060760857405080101A203020101A305A103020101BE0604040E010601"
        )
        aare = acse.ApplicationAssociationResponse.from_bytes(data)
        assert not aare.ciphered
        assert aare.result == enumerations.AssociationResult.REJECTED_PERMANENT
        assert (
            aare.result_source_diagnostics
            == enumerations.AcseServiceUserDiagnostics.NO_REASON_GIVEN
        )
        print(aare.user_information)
        assert aare.user_information.content is not None
        assert (
            aare.user_information.content.error
            == enumerations.InitiateError.DLMS_VERSION_TOO_LOW
        )

    def test_no_ciher_hls_ok(self):
        data = bytes.fromhex(
            "6142A109060760857405080101A203020100A305A10302010E88020780890760857405080205AA0A8008503677524A323146BE10040E0800065F1F040000501F01F40007"
        )
        aare = acse.ApplicationAssociationResponse.from_bytes(data)
        assert not aare.ciphered
        assert aare.result == enumerations.AssociationResult.ACCEPTED
        assert (
            aare.result_source_diagnostics
            == enumerations.AcseServiceUserDiagnostics.AUTHENTICATION_REQUIRED
        )
        assert aare.authentication_value is not None
        assert aare.authentication == enumerations.AuthenticationMechanism.HLS_GMAC
        assert isinstance(aare.user_information.content, InitiateResponse)

    def test_no_cipher_no_auth_conformance(self):
        data = bytes.fromhex(
            "6129a109060760857405080101a203020100a305a103020100be10040e0800065f1f0400001e1d04c80007"
        )

        aare = acse.ApplicationAssociationResponse.from_bytes(data)

        assert aare.result == enumerations.AssociationResult.ACCEPTED
        assert (
            aare.result_source_diagnostics
            == enumerations.AcseServiceUserDiagnostics.NULL
        )
        assert not aare.ciphered
        assert not aare.authentication
        assert aare.system_title is None
        assert aare.authentication_value is None
        assert aare.public_cert is None
        assert isinstance(aare.user_information.content, InitiateResponse)
        assert aare.user_information.content.negotiated_conformance == Conformance(
            general_protection=False,
            general_block_transfer=False,
            delta_value_encoding=False,
            attribute_0_supported_with_set=False,
            priority_management_supported=False,
            attribute_0_supported_with_get=False,
            block_transfer_with_get_or_read=True,
            block_transfer_with_set_or_write=True,
            block_transfer_with_action=True,
            multiple_references=True,
            data_notification=False,
            access=False,
            get=True,
            set=True,
            selective_access=True,
            event_notification=False,
            action=True,
        )


class TestEncodeAARE:
    def test_no_ciphering_no_security_ok_association(self):
        data = bytes.fromhex(
            "6129A109060760857405080101A203020100A305A103020100BE10040E0800065F1F040000501F01F40007"
        )
        aare = acse.ApplicationAssociationResponse(
            result=enumerations.AssociationResult.ACCEPTED,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.NULL,
            ciphered=False,
            authentication=None,
            system_title=None,
            public_cert=None,
            authentication_value=None,
            user_information=acse.UserInformation(
                content=InitiateResponse(
                    negotiated_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=False,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=False,
                        block_transfer_with_action=False,
                        multiple_references=False,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    server_max_receive_pdu_size=500,
                    negotiated_dlms_version_number=6,
                    negotiated_quality_of_service=0,
                )
            ),
            implementation_information=None,
            responding_ap_invocation_id=None,
            responding_ae_invocation_id=None,
        )

        assert aare.to_bytes() == data

    def test_no_ciphering_no_security_app_context_name_wrong(self):
        data = bytes.fromhex(
            "6129A109060760857405080101A203020101A305A103020102BE10040E0800065F1F040000501F01F40007"
        )
        aare = acse.ApplicationAssociationResponse(
            result=enumerations.AssociationResult.REJECTED_PERMANENT,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.APPLICATION_CONTEXT_NAME_NOT_SUPPORTED,
            ciphered=False,
            authentication=None,
            system_title=None,
            public_cert=None,
            authentication_value=None,
            user_information=acse.UserInformation(
                content=InitiateResponse(
                    negotiated_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=False,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=False,
                        block_transfer_with_action=False,
                        multiple_references=False,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    server_max_receive_pdu_size=500,
                    negotiated_dlms_version_number=6,
                    negotiated_quality_of_service=0,
                )
            ),
            implementation_information=None,
            responding_ap_invocation_id=None,
            responding_ae_invocation_id=None,
        )

        assert aare.to_bytes() == data

    def test_no_cipher_no_security_incorrect_dlms_version(self):
        data = bytes.fromhex(
            "611FA109060760857405080101A203020101A305A103020101BE0604040E010601"
        )

        aare = acse.ApplicationAssociationResponse(
            result=enumerations.AssociationResult.REJECTED_PERMANENT,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.NO_REASON_GIVEN,
            ciphered=False,
            authentication=None,
            system_title=None,
            public_cert=None,
            authentication_value=None,
            user_information=acse.UserInformation(
                content=ConfirmedServiceError(
                    error=enumerations.InitiateError.DLMS_VERSION_TOO_LOW
                )
            ),
            implementation_information=None,
            responding_ap_invocation_id=None,
            responding_ae_invocation_id=None,
        )

        assert aare.to_bytes() == data

    def test_no_ciher_hls_ok(self):
        data = bytes.fromhex(
            "6142A109060760857405080101A203020100A305A10302010E88020780890760857405080205AA0A8008503677524A323146BE10040E0800065F1F040000501F01F40007"
        )

        aare = acse.ApplicationAssociationResponse(
            result=enumerations.AssociationResult.ACCEPTED,
            result_source_diagnostics=enumerations.AcseServiceUserDiagnostics.AUTHENTICATION_REQUIRED,
            ciphered=False,
            authentication=enumerations.AuthenticationMechanism.HLS_GMAC,
            public_cert=None,
            system_title=None,
            authentication_value=b"P6wRJ21F",
            user_information=UserInformation(
                content=InitiateResponse(
                    negotiated_conformance=Conformance(
                        general_protection=False,
                        general_block_transfer=False,
                        delta_value_encoding=False,
                        attribute_0_supported_with_set=False,
                        priority_management_supported=True,
                        attribute_0_supported_with_get=False,
                        block_transfer_with_get_or_read=True,
                        block_transfer_with_set_or_write=False,
                        block_transfer_with_action=False,
                        multiple_references=False,
                        data_notification=False,
                        access=False,
                        get=True,
                        set=True,
                        selective_access=True,
                        event_notification=True,
                        action=True,
                    ),
                    server_max_receive_pdu_size=500,
                    negotiated_dlms_version_number=6,
                    negotiated_quality_of_service=0,
                )
            ),
            implementation_information=None,
            responding_ap_invocation_id=None,
            responding_ae_invocation_id=None,
        )

        assert aare.to_bytes() == data



import contextlib
from typing import *

import attr
import structlog

from dlms_cosem import cosem, dlms_data, enumerations, exceptions, state, utils
from dlms_cosem.security import AuthenticationMethodManager
from dlms_cosem.io import DlmsTransport
from dlms_cosem.connection import DlmsConnection, DlmsConnectionSettings
from dlms_cosem.cosem.selective_access import RangeDescriptor
from dlms_cosem.protocol import acse, xdlms
from dlms_cosem.protocol.xdlms import ConfirmedServiceError

LOG = structlog.get_logger()


class DataResultError(Exception):
    """ Error retrieveing data"""


class ActionError(Exception):
    """Error performing an action"""


class HLSError(Exception):
    """error in HLS procedure"""


@attr.s(auto_attribs=True)
class DlmsClient:
    transport: DlmsTransport
    authentication: AuthenticationMethodManager
    encryption_key: Optional[bytes] = attr.ib(default=None)
    authentication_key: Optional[bytes] = attr.ib(default=None)
    security_suite: Optional[int] = attr.ib(default=0)
    dedicated_ciphering: bool = attr.ib(default=False)
    block_transfer: bool = attr.ib(default=False)
    max_pdu_size: int = attr.ib(default=65535)
    client_system_title: Optional[bytes] = attr.ib(default=None)
    client_initial_invocation_counter: int = attr.ib(default=0)
    meter_initial_invocation_counter: int = attr.ib(default=0)
    timeout: int = attr.ib(default=10)
    connection_settings: Optional[DlmsConnectionSettings] = attr.ib(default=None)

    dlms_connection: DlmsConnection = attr.ib(
        default=attr.Factory(
            lambda self: DlmsConnection(
                client_system_title=self.client_system_title,
                authentication=self.authentication,
                global_encryption_key=self.encryption_key,
                global_authentication_key=self.authentication_key,
                use_dedicated_ciphering=self.dedicated_ciphering,
                use_block_transfer=self.block_transfer,
                security_suite=self.security_suite,
                max_pdu_size=self.max_pdu_size,
                client_invocation_counter=self.client_initial_invocation_counter,
                meter_invocation_counter=self.meter_initial_invocation_counter,
                settings=self.connection_settings,
            ),
            takes_self=True,
        )
    )

    @contextlib.contextmanager
    def session(self) -> "DlmsClient":
        self.connect()
        self.associate()
        yield self
        self.release_association()
        self.disconnect()

    def get(
            self,
            cosem_attribute: cosem.CosemAttribute,
            access_descriptor: Optional[RangeDescriptor] = None,
    ) -> bytes:
        self.send(
            xdlms.GetRequestNormal(
                cosem_attribute=cosem_attribute, access_selection=access_descriptor
            )
        )
        all_data_received = False
        data = bytearray()
        while not all_data_received:
            get_response = self.next_event()
            if isinstance(get_response, xdlms.GetResponseNormal):
                data.extend(get_response.data)
                all_data_received = True
                continue
            if isinstance(get_response, xdlms.GetResponseWithBlock):
                data.extend(get_response.data)
                self.send(
                    xdlms.GetRequestNext(
                        invoke_id_and_priority=get_response.invoke_id_and_priority,
                        block_number=get_response.block_number,
                    )
                )
                continue
            if isinstance(get_response, xdlms.GetResponseLastBlock):
                data.extend(get_response.data)
                all_data_received = True
                continue

            if isinstance(get_response, xdlms.GetResponseLastBlockWithError):
                raise DataResultError(
                    f"Error in blocktransfer of GET response: {get_response.error!r}"
                )

            if isinstance(get_response, xdlms.GetResponseNormalWithError):
                raise DataResultError(
                    f"Could not perform GET request: {get_response.error!r}"
                )

        return bytes(data)

    def get_many(
            self, cosem_attributes_with_selection: List[cosem.CosemAttributeWithSelection]
    ):
        """
        Make a GET.WITH_LIST call. Get many items in one request.
        """
        out = xdlms.GetRequestWithList(
            cosem_attributes_with_selection=cosem_attributes_with_selection
        )
        self.send(out)
        response = self.next_event()
        if isinstance(response, xdlms.ExceptionResponse):
            raise exceptions.DlmsClientException(
                f"Received an Exception response with state error: "
                f"{response.state_error.name} and service error: "
                f"{response.service_error.name}"
            )
        return response

    def set(self, cosem_attribute: cosem.CosemAttribute, data: bytes):
        self.send(xdlms.SetRequestNormal(cosem_attribute=cosem_attribute, data=data))
        return self.next_event()

    def action(self, method: cosem.CosemMethod, data: bytes):
        self.send(xdlms.ActionRequestNormal(cosem_method=method, data=data))
        response = self.next_event()

        if isinstance(response, xdlms.ActionResponseNormalWithError):
            raise ActionError(response.error.name)
        elif isinstance(response, xdlms.ActionResponseNormalWithData):
            if response.status != enumerations.ActionResultStatus.SUCCESS:
                raise ActionError(f"Unsuccessful ActionRequest: {response.status.name}")
            return response.data
        else:
            if response.status != enumerations.ActionResultStatus.SUCCESS:
                raise ActionError(f"Unsuccessful ActionRequest: {response.status.name}")
        return

    def associate(
            self,
            association_request: Optional[acse.ApplicationAssociationRequest] = None,
    ) -> acse.ApplicationAssociationResponse:

        # the aarq can be overridden or the standard one from the connection is used.
        aarq = association_request or self.dlms_connection.get_aarq()

        self.send(aarq)
        response = self.next_event()
        # we could have received an exception from the meter.
        if isinstance(response, xdlms.ExceptionResponse):
            raise exceptions.DlmsClientException(
                f"DLMS Exception: {response.state_error!r}:{response.service_error!r}"
            )
        # the association might not be accepted by the meter
        if isinstance(response, acse.ApplicationAssociationResponse):
            if response.result is not enumerations.AssociationResult.ACCEPTED:
                # there could be an error suppled with the reject.
                extra_error = None
                if response.user_information:
                    if isinstance(
                            response.user_information.content, ConfirmedServiceError
                    ):
                        extra_error = response.user_information.content.error
                raise exceptions.DlmsClientException(
                    f"Unable to perform Association: {response.result!r} and "
                    f"{response.result_source_diagnostics!r}, extra info: {extra_error}"
                )
        else:
            raise exceptions.LocalDlmsProtocolError(
                "Did not receive an AARE after sending AARQ"
            )

        if self.should_send_hls_reply():

            # TODO: wrap hls logic in method
            try:
                hls_response = self.send_hls_reply()
            except ActionError as e:
                raise HLSError from e

            if not hls_response:
                raise HLSError("No HLS data in response")

            hls_data = utils.parse_as_dlms_data(hls_response)

            if not hls_data:
                raise HLSError("Did not receive any HLS response data")

            if not self.dlms_connection.authentication.hls_meter_data_is_valid(
                    hls_data, self.dlms_connection
            ):
                raise HLSError(
                    f"Meter did not respond with correct challenge calculation"
                )

        return response

    def should_send_hls_reply(self) -> bool:
        return (
                self.dlms_connection.state.current_state
                == state.SHOULD_SEND_HLS_SEVER_CHALLENGE_RESULT
        )

    def send_hls_reply(self) -> Optional[bytes]:
        return self.action(
            method=cosem.CosemMethod(
                enumerations.CosemInterface.ASSOCIATION_LN,
                cosem.Obis(0, 0, 40, 0, 0),
                1,
            ),
            data=dlms_data.OctetStringData(
                self.dlms_connection.authentication.hls_generate_reply_data(
                    self.dlms_connection
                )
            ).to_bytes(),
        )

    def release_association(self) -> Optional[acse.ReleaseResponse]:

        rlrq = self.dlms_connection.get_rlrq()
        try:
            self.send(rlrq)
            rlre = self.next_event()
            return rlre
        except exceptions.NoRlrqRlreError:
            return None

    def connect(self):
        self.transport.connect()

    def disconnect(self):
        self.transport.disconnect()

    def send(self, *events):
        for event in events:
            data = self.dlms_connection.send(event)
            response_bytes = self.transport.send_request(data)
            self.dlms_connection.receive_data(response_bytes)

    def next_event(self):
        event = self.dlms_connection.next_event()
        return event

    @property
    def client_invocation_counter(self) -> int:
        return self.dlms_connection.client_invocation_counter

    @client_invocation_counter.setter
    def client_invocation_counter(self, ic: int):
        self.dlms_connection.client_invocation_counter = ic



from functools import partial
from typing import Optional

import attr

from dlms_cosem import a_xdr
from dlms_cosem.dlms_data import OctetStringData
from dlms_cosem.protocol.xdlms.base import AbstractXDlmsApdu
from dlms_cosem.security import SecurityControlField, decrypt

int_from_bytes = partial(int.from_bytes, "big")


@attr.s(auto_attribs=True)
class GeneralGlobalCipher(AbstractXDlmsApdu):
    """
    The general-global-cipher APDU can be used to cipher other APDUs with
    either the global key or the dedicated key.

    The additional authenticated data to use for decryption is depending on the
    portection applied.

    Encrypted and authenticated: Security Control Field || Authentication Key
    Only authenticated: Security Control Field || Authentication Key || Ciphered Text
    Only encrypted: b''
    No protection: b''

    """

    TAG = 219
    NAME = "general-glo-cipher"

    ENCODING_CONF = a_xdr.EncodingConf(
        [
            a_xdr.Attribute(
                attribute_name="system_title", create_instance=OctetStringData
            ),
            a_xdr.Attribute(
                attribute_name="ciphered_content",
                create_instance=OctetStringData.from_bytes,
            ),
        ]
    )

    # Some implementations does not send the system_title. But it seems like it is against the standard.
    system_title: Optional[bytes]
    security_control: SecurityControlField
    invocation_counter: int
    ciphered_text: bytes

    @classmethod
    def from_bytes(cls, source_bytes: bytes):
        data = bytearray(source_bytes)
        tag = data.pop(0)
        if tag != cls.TAG:
            raise ValueError(f"Tag not as expected. Expected: {cls.TAG} but got {tag}")
        decoder = a_xdr.AXdrDecoder(encoding_conf=cls.ENCODING_CONF)
        in_dict = decoder.decode(data)
        system_title = in_dict["system_title"].value
        ciphered_content = in_dict["ciphered_content"].value
        security_control = SecurityControlField.from_bytes(
            ciphered_content.pop(0).to_bytes(1, "big")
        )
        invocation_counter = int.from_bytes(ciphered_content[:4], "big")
        ciphered_text = bytes(ciphered_content[4:])
        return cls(system_title, security_control, invocation_counter, ciphered_text)

    def to_bytes(self) -> bytes:
        out = bytearray()
        out.append(self.TAG)
        if self.system_title:
            out.append(len(self.system_title))
            out.extend(self.system_title)
        else:
            out.extend(b"\x00")
        out.append(
            len(
                self.security_control.to_bytes()
                + self.invocation_counter.to_bytes(4, "big")
                + self.ciphered_text
            )
        )
        out.extend(self.security_control.to_bytes())
        out.extend(self.invocation_counter.to_bytes(4, "big"))
        out.extend(self.ciphered_text)
        return bytes(out)

    def to_plain_apdu(self, encryption_key, authentication_key) -> bytes:
        plain_text = decrypt(
            security_control=self.security_control,
            key=encryption_key,
            auth_key=authentication_key,
            invocation_counter=self.invocation_counter,
            cipher_text=self.ciphered_text,
            system_title=self.system_title,
        )

        return bytes(plain_text)
